import sympy as sym
import numpy as np

def dcdc_ph(grid,vsc_data):
    '''
    dc/dc converter grid former in the low voltage side
    
    '''
    
    bus_hv_name = vsc_data['bus_hv']
    bus_lv_name = vsc_data['bus_lv']  
    
    A_value  = vsc_data['A']   
    B_value  = vsc_data['B']   
    C_value  = vsc_data['C']  

    #### voltages:
    v_hp, v_hn = sym.symbols(f'V_{bus_hv_name}_0_r,V_{bus_hv_name}_1_r', real=True)
    v_lp, v_ln = sym.symbols(f'V_{bus_lv_name}_0_r,V_{bus_lv_name}_1_r', real=True)
    v_hp_i, v_hn_i = sym.symbols(f'V_{bus_hv_name}_0_i,V_{bus_hv_name}_1_i', real=True)
   
    ### LV-side
    i_l,p_l = sym.symbols(f'i_l_{bus_lv_name},p_l_{bus_lv_name}',real=True)
    A,B,C = sym.symbols(f'A_{bus_hv_name},B_{bus_hv_name},C_{bus_hv_name}',real=True)

    ### HV-side
    e_h,v_h,i_h,p_h = sym.symbols(f'e_h_{bus_hv_name},v_h_{bus_hv_name},i_h_{bus_hv_name},p_h_{bus_hv_name}',real=True)
    R_h = sym.Symbol(f'R_h_{bus_hv_name}',real=True)
    p_h_ref = sym.Symbol(f'p_ref_{bus_hv_name}',real=True)
    m_h = sym.Symbol(f'm_h_{bus_hv_name}',real=True)
    i_g = sym.Symbol(f'i_g_{bus_hv_name}',real=True) 
    R_g = sym.Symbol(f'R_g_{bus_hv_name}',real=True)



    v_l = v_lp-v_ln
    p_h = p_h_ref
    v_h = v_hp - v_hn
    i_h = p_h/v_h
    i_hp =  i_h
    i_hn = -i_h
    p_loss = A*i_hp**2 + B*sym.Abs(i_hp) + C
    p_l = p_h + p_loss
    i_l =  p_l/v_l  

    # current injections dc HV side
    idx_lp_r,imag = grid.node2idx(f'{bus_lv_name}','a')
    idx_ln_r,imag = grid.node2idx(f'{bus_lv_name}','b')
    grid.dae['g'] [idx_lp_r] +=  i_l
    grid.dae['g'] [idx_ln_r] += -i_l

    # current injections dc LV side
    idx_hp_r,idx_hp_i = grid.node2idx(f'{bus_hv_name}','a')
    idx_hn_r,idx_hn_i = grid.node2idx(f'{bus_hv_name}','b')
    grid.dae['g'] [idx_hp_r] += -i_hp 
    grid.dae['g'] [idx_hn_r] += -i_hn
    grid.dae['g'] [idx_hp_i] += -v_hp_i/1e3
    grid.dae['g'] [idx_hn_i] += -v_hn_i/1e3


    grid.dae['u_ini_dict'].update({f'p_ref_{bus_hv_name}':vsc_data['p_ref']}) 
    grid.dae['u_run_dict'].update({f'p_ref_{bus_hv_name}':vsc_data['p_ref']}) 
   
    grid.dae['params_dict'].update({f'R_h_{bus_hv_name}':0.01})    
    grid.dae['params_dict'].update({f'R_g_{bus_hv_name}':3})   
    grid.dae['params_dict'].update({f'A_{bus_hv_name}':A_value})   
    grid.dae['params_dict'].update({f'B_{bus_hv_name}':B_value})   
    grid.dae['params_dict'].update({f'C_{bus_hv_name}':C_value})   

    grid.dae['h_dict'].update({f'p_loss_{bus_hv_name}':p_loss})


def test():
    import numpy as np
    import sympy as sym
    import json
    from pydae.urisi.urisi_builder import urisi
    import pydae.build_cffi as db

    grid = urisi('dcdc_ph.hjson')
    grid.uz_jacs = True
    grid.construct('temp')
    grid.compile('temp')

    import temp

    S_n = 100e3
    V_n = 1000
    I_n = S_n/V_n
    Conduction_losses = 0.02*S_n # = A*I_n**2
    A = Conduction_losses/(I_n**2)
    B = 1
    C = 0.02*S_n
    model = temp.model()
    model.ini({'A_D2':A,'B_D2':B,'C_D2':C},'xy_0.json')
    model.report_y()
    model.report_z()



if __name__ == '__main__':

    #development()
    test()






import numpy as np
import matplotlib.pyplot as plt
import ipywidgets as widgets
import pydae.svg_tools as st 
import pydae.build_cffi as db
from pydae.bmapu.pvs.utils.pv_builder import pv_model

plt.style.use('https://raw.githubusercontent.com/pydae/pydae/master/src/pydae/edashboards/presentation.mplstyle')
colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
plt.rcParams['font.family'] =  ['sans-serif']

class dashboard():
    
    def __init__(self):

        
        self.colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
        self.data = {"I_sc":3.87,"V_oc":42.1,"I_mpp":3.56,"V_mpp":33.7,"N_s":72,
        "K_vt":-0.160,"K_it":0.065,"R_pv_s": 0.5602, "R_pv_sh": 1862, "K_d": 1.3433}


    def ini(self):

        import pv_test

        self.model = pv_test.model()
        self.model.Dt = 0.01
        self.model.decimation = 1
        self.model.ini(self.data,{'i_pv':3,'v_mpp':40})
        self.plot_V_max = self.data['V_oc']*1.2
        self.plot_P_max = self.data['V_mpp']*self.data['I_mpp']*1.2
        self.plot_I_max = self.data['I_sc']*1.2
        plt.ioff()
        plt.clf()
        
    def build(self):

        pv = pv_model('pv_test')
        pv.model_params(self.data)
        pv.build()

        self.pv = pv

    def compute_curve(self,V):

        self.I = np.zeros(len(self.V))
    
        irrad = self.sld_irrad.value
        temp_deg = self.sld_temp_deg.value

        for it,v in enumerate(self.V):
            self.model.ini({'v_pv':v,'irrad':irrad,'temp_deg':temp_deg})
            self.I[it] = self.model.get_value('i_pv')

    def compute_point(self,v):
        irrad = self.sld_irrad.value
        temp_deg = self.sld_temp_deg.value
        self.v_pv = v
        self.model.ini({'v_pv':v,'irrad':irrad,'temp_deg':temp_deg})
        self.i_pv  = self.model.get_value('i_pv')


    def widgets(self):
        
        model = self.model
        colors = self.colors

        plt.ioff()
        plt.clf()

        self.sld_irrad = widgets.FloatSlider(description='irrad',min=0, max=1200, step=50, value=1000)
        self.sld_temp_deg = widgets.FloatSlider(description='temp (deg)',min=0, max=75, step=1, value=25)
        self.sld_v_pv = widgets.FloatSlider(description='v<sub>pv</sub>*',min=0.0, max=self.data['V_oc']*1.2, step=1, value=33)
        
        self.V = np.arange(0,50,1.0)
        self.compute_curve(self.V)

        fig,axes = plt.subplots(ncols=2,figsize=(6,3));

        self.curve_vi = axes[0].plot(self.V,self.I, color=colors[0]);
        self.curve_vp = axes[1].plot(self.V,self.V*self.I, color=colors[1]);

        self.compute_point(self.sld_v_pv.value)

        #self.curve_vi_point = axes[0].plot(self.V,self.I,'o', color='k');
        # self.curve_vp_point = axes[1].plot(self.v_pv,self.v_pv*self.i_pv,'o', color='k');

        axes[0].set_ylim(0,self.plot_I_max)
        axes[0].set_xlim(0,self.plot_V_max)  
         
        axes[1].set_ylim(0,self.plot_P_max)
        axes[1].set_xlim(0,self.plot_V_max)    
        axes[1].set_xlim(0,self.plot_V_max)    

        axes[0].set_xlabel(r'$\sf v_{pv}$ (V)')
        axes[1].set_xlabel(r'$\sf v_{pv}$ (V)')
        axes[0].set_ylabel(r'$\sf i_{pv}$ (A)')
        axes[1].set_ylabel(r'$\sf p_{pv}$ (W)')

        fig.savefig('curve_iv_pv.svg')

        self.s_pv_fig = st.svg(r"pv_model_db.svg")
        self.html_pv_fig = widgets.HTML(
            value= self.s_pv_fig.tostring(),
            placeholder='',
            description='',
        )

        self.s_curves = st.svg(r"curve_iv_pv.svg")
        self.html_curves = widgets.HTML(
            value= self.s_curves.tostring(),
            placeholder='',
            description='',
        )

        self.fig = fig
        self.axes = axes

        
    def update(self,change):
        
        self.V = np.arange(0,self.data['V_oc']*1.25,1.0)
        self.compute_curve(self.V)

        self.curve_vi[0].set_data(self.V,self.I)
        self.curve_vp[0].set_data(self.V,self.V*self.I)

        self.compute_point(self.sld_v_pv.value)

        #self.curve_vi_point.set_data(self.V,self.I+0.1)
        # self.curve_vp_point[0].set_data(self.v_pv,self.v_pv*self.i_pv)


        self.fig.savefig('curve_iv_pv.svg')
        self.s_curves = st.svg(r"curve_iv_pv.svg")

        self.html_curves.value = self.s_curves.tostring()

        self.s_pv_fig.set_tspan('v_pv', f'={self.model.get_value("v_pv"):4.1f} V')
        self.s_pv_fig.set_tspan('i_pv', f'={self.model.get_value("i_pv"):4.1f} A')
        self.s_pv_fig.set_tspan('p_pv', f'={self.model.get_value("v_pv")*self.model.get_value("i_pv"):5.1f} W')

        self.html_pv_fig.value = self.s_pv_fig.tostring()

    def show(self):

        # Link the sliders value to the update function
        self.sld_irrad.observe(self.update, names='value')
        self.sld_temp_deg.observe(self.update, names='value')
        self.sld_v_pv.observe(self.update, names='value')

        self.update(0)
        self.layout = widgets.HBox([widgets.VBox([self.html_pv_fig,
                                                  self.sld_irrad,
                                                  self.sld_temp_deg,
                                                  self.sld_v_pv]),
                                    widgets.VBox([self.html_curves])])
                                    
        display(self.layout)


import sympy as sym
import numpy as np

def dcdc_gfl(grid,vsc_data):
    '''
    dc/dc converter grid former in the low voltage side
    
    '''
    
    bus_hv_name = vsc_data['bus_hv']
    bus_lv_name = vsc_data['bus_lv']  
    
    A_value  = vsc_data['A']   
    B_value  = vsc_data['B']   
    C_value  = vsc_data['C']   

    #### voltages:
    v_hp, v_hn = sym.symbols(f'V_{bus_hv_name}_0_r,V_{bus_hv_name}_1_r', real=True)
    v_lp, v_ln = sym.symbols(f'V_{bus_lv_name}_0_r,V_{bus_lv_name}_1_r', real=True)
    v_hp_i, v_hn_i = sym.symbols(f'V_{bus_hv_name}_0_i,V_{bus_hv_name}_1_i', real=True)
    v_lp_i, v_ln_i = sym.symbols(f'V_{bus_lv_name}_0_i,V_{bus_lv_name}_1_i', real=True)

    ### HV-side
    e_h,v_h,i_h,p_h = sym.symbols(f'e_h_{bus_hv_name},v_h_{bus_hv_name},i_h_{bus_hv_name},p_h_{bus_hv_name}',real=True)
    R_h = sym.Symbol(f'R_h_{bus_hv_name}',real=True)

    ### LV-side
    e_l,v_l,i_l,p_l = sym.symbols(f'e_l_{bus_lv_name},v_l_{bus_lv_name},i_l_{bus_lv_name},p_l_{bus_lv_name}',real=True)
    A,B,C = sym.symbols(f'A_{bus_lv_name},B_{bus_lv_name},C_{bus_lv_name}',real=True)
    R_l = sym.Symbol(f'R_l_{bus_lv_name}',real=True)
    e_l_ref = sym.Symbol(f'e_l_ref_{bus_lv_name}',real=True)
    m_l = sym.Symbol(f'm_l_{bus_lv_name}',real=True)
    R_g = sym.Symbol(f'R_g_{bus_lv_name}',real=True)

    v_h = v_hp-v_hn
    m_l = e_l_ref/v_h
    e_l = e_l_ref
    i_lp = (v_ln + e_l - v_lp)/R_l 
    i_ln = -i_lp-v_ln/R_g
    p_l = i_lp*e_l
    p_h = p_l + A*i_lp**2 + B*sym.Abs(i_lp) + C
    i_h =  p_h/v_h 

    # current injections dc HV side
    idx_hp_r,imag = grid.node2idx(f'{bus_hv_name}','a')
    idx_hn_r,imag = grid.node2idx(f'{bus_hv_name}','b')
    grid.dae['g'] [idx_hp_r] +=  i_h
    grid.dae['g'] [idx_hn_r] += -i_h

    # current injections dc LV side
    idx_lp_r,idx_lp_i = grid.node2idx(f'{bus_lv_name}','a')
    idx_ln_r,idx_ln_i = grid.node2idx(f'{bus_lv_name}','b')
    grid.dae['g'] [idx_lp_r] += -i_lp 
    grid.dae['g'] [idx_ln_r] += -i_ln
    grid.dae['g'] [idx_lp_i] += -v_lp_i/1e3
    grid.dae['g'] [idx_ln_i] += -v_ln_i/1e3

    grid.dae['u_ini_dict'].update({f'e_l_ref_{bus_lv_name}':400.0}) 
    grid.dae['u_run_dict'].update({f'e_l_ref_{bus_lv_name}':400.0}) 
   
    grid.dae['params_dict'].update({f'R_l_{bus_lv_name}':0.1})    
    grid.dae['params_dict'].update({f'R_g_{bus_lv_name}':3})   
    grid.dae['params_dict'].update({f'A_{bus_lv_name}':A_value})   
    grid.dae['params_dict'].update({f'B_{bus_lv_name}':B_value})   
    grid.dae['params_dict'].update({f'C_{bus_lv_name}':C_value})   

  #  grid.dae['xy_0_dict'].update({f'i_g_{bus_lv_name}':0.0})
    
    grid.dae['h_dict'].update({f'm_l_{bus_lv_name}':m_l})
   # grid.dae['h_dict'].update({f'i_g_{bus_lv_name}':i_g})


if __name__ == '__main__':

    import numpy as np
    import sympy as sym
    import json
    from pydae.urisi.urisi_builder import urisi
    import pydae.build_cffi as db

    grid = urisi('dcdc_gfl.hjson')
    grid.uz_jacs = True
    grid.construct('temp')
    grid.compile('temp')

    import temp

    S_n = 100e3
    V_n = 400
    I_n = S_n/V_n
    Conduction_losses = 0.02*S_n # = A*I_n**2
    A = Conduction_losses/(I_n**2)
    B = 1
    C = 0.02*S_n
    model = temp.model()
    model.ini({'A_D2':A,'B_D2':B,'C_D2':C},'xy_0.json')
    model.report_y()
    model.report_z()
    #model.save_xy_0('xy_1.json')



import sympy as sym
import numpy as np

def acdc_3ph_4w_pq(grid,vsc_data):
    '''
    Converter type p_ac,q_ac 3 phase 4 wire
    
    '''
    
    bus_ac_name = vsc_data['bus_ac']
    bus_dc_name = vsc_data['bus_dc']  
    
    a_value  = vsc_data['A']   
    b_value  = vsc_data['B']   
    c_value  = vsc_data['C']   
    
    ### AC-side
    p_a,p_b,p_c = sym.symbols(f'p_vsc_a_{bus_ac_name},p_vsc_b_{bus_ac_name},p_vsc_c_{bus_ac_name}',real=True)
    q_a,q_b,q_c = sym.symbols(f'q_vsc_a_{bus_ac_name},q_vsc_b_{bus_ac_name},q_vsc_c_{bus_ac_name}',real=True)

    p_ac,q_ac,p_dc,p_loss = sym.symbols(f'p_vsc_{bus_ac_name},q_vsc_{bus_ac_name},p_vsc_{bus_dc_name},p_vsc_loss_{bus_ac_name}',real=True)
    p_a_d,p_b_d,p_c_d,p_n_d = sym.symbols(f'p_a_d_{bus_ac_name},p_b_d_{bus_ac_name},p_c_d_{bus_ac_name},p_n_d_{bus_ac_name}',real=True)
   
    #### AC voltages:
    v_a_r,v_b_r,v_c_r,v_n_r = sym.symbols(f'V_{bus_ac_name}_0_r,V_{bus_ac_name}_1_r,V_{bus_ac_name}_2_r,V_{bus_ac_name}_3_r', real=True)
    v_a_i,v_b_i,v_c_i,v_n_i = sym.symbols(f'V_{bus_ac_name}_0_i,V_{bus_ac_name}_1_i,V_{bus_ac_name}_2_i,V_{bus_ac_name}_3_i', real=True)
   
    #### AC currents:
    i_a_r,i_a_i = sym.symbols(f'i_vsc_{bus_ac_name}_a_r,i_vsc_{bus_ac_name}_a_i',real=True)
    i_b_r,i_b_i = sym.symbols(f'i_vsc_{bus_ac_name}_b_r,i_vsc_{bus_ac_name}_b_i',real=True)
    i_c_r,i_c_i = sym.symbols(f'i_vsc_{bus_ac_name}_c_r,i_vsc_{bus_ac_name}_c_i',real=True)
    i_n_r,i_n_i = sym.symbols(f'i_vsc_{bus_ac_name}_n_r,i_vsc_{bus_ac_name}_n_i',real=True)
    
    #### DC voltages:
    v_pos,v_neg = sym.symbols(f'V_{bus_dc_name}_0_r,V_{bus_dc_name}_1_r', real=True)
    v_posi,v_negi = sym.symbols(f'V_{bus_dc_name}_0_i,V_{bus_dc_name}_1_i', real=True)

    # algebraic states dc side
    i_pos,i_neg,v_og = sym.symbols(f'i_vsc_pos_{bus_dc_name}_sp,i_vsc_{bus_dc_name}_sn, v_og_{bus_dc_name}', real=True)

    A_loss,B_loss,C_loss = sym.symbols(f'A_loss_{bus_ac_name},B_loss_{bus_ac_name},C_loss_{bus_ac_name}',real=True)
    R_dc,K_dc,R_gdc = sym.symbols(f'R_dc_{bus_dc_name},K_dc_{bus_dc_name},R_gdc_{bus_dc_name}',real=True)
     

    i_a_rms = sym.sqrt(i_a_r**2+i_a_i**2 + 0.01) 
    i_b_rms = sym.sqrt(i_b_r**2+i_b_i**2+ 0.01) 
    i_c_rms = sym.sqrt(i_c_r**2+i_c_i**2+ 0.01) 
    i_n_rms = sym.sqrt(i_n_r**2+i_n_i**2+ 0.01) 

    p_loss_a = A_loss + B_loss*i_a_rms + C_loss*i_a_rms*i_a_rms
    p_loss_b = A_loss + B_loss*i_b_rms + C_loss*i_b_rms*i_b_rms
    p_loss_c = A_loss + B_loss*i_c_rms + C_loss*i_c_rms*i_c_rms
    p_loss_n = A_loss + B_loss*i_n_rms + C_loss*i_n_rms*i_n_rms

    v_a = v_a_r + 1j*v_a_i
    v_b = v_b_r + 1j*v_b_i
    v_c = v_c_r + 1j*v_c_i
    v_n = v_n_r + 1j*v_n_i

    i_a = i_a_r + 1j*i_a_i
    i_b = i_b_r + 1j*i_b_i
    i_c = i_c_r + 1j*i_c_i
    i_n = i_n_r + 1j*i_n_i

    s_a = (v_a - v_n) * sym.conjugate(i_a)
    s_b = (v_b - v_n) * sym.conjugate(i_b)
    s_c = (v_c - v_n) * sym.conjugate(i_c)
    s_n = (v_n) * sym.conjugate(i_n)   
   
    eq_i_a_r =  sym.re(s_a) - p_a
    eq_i_b_r =  sym.re(s_b) - p_b
    eq_i_c_r =  sym.re(s_c) - p_c
    eq_i_a_i =  sym.im(s_a) - q_a
    eq_i_b_i =  sym.im(s_b) - q_b
    eq_i_c_i =  sym.im(s_c) - q_c
    
    eq_i_n_r = i_n_r + i_a_r + i_b_r + i_c_r
    eq_i_n_i = i_n_i + i_a_i + i_b_i + i_c_i
   
    v_dc = v_pos - v_neg
    
    p_ac = p_a + p_b + p_c
    p_loss_total = p_loss_a+p_loss_b+p_loss_c+p_loss_n 

    eq_p_dc = -p_dc +  p_ac + p_loss_total

    eq_i_pos = -i_pos + p_dc/v_dc
    eq_i_neg = -i_neg - p_dc/v_dc
    #eq_v_og = -v_og/R_gdc - i_pos - i_neg 


    grid.dae['g'] += [eq_i_a_r,eq_i_a_i,
                      eq_i_b_r,eq_i_b_i,
                      eq_i_c_r,eq_i_c_i,
                      eq_i_n_r,eq_i_n_i,
                      eq_i_pos,eq_i_neg,#eq_v_og,
                      eq_p_dc
                      ]
    
    grid.dae['y_ini'] += [i_a_r,   i_a_i,
                      i_b_r,   i_b_i,
                      i_c_r,   i_c_i,
                      i_n_r,   i_n_i,
                      i_pos,i_neg,#v_og,
                      p_dc]

    grid.dae['y_run'] += [i_a_r,   i_a_i,
                      i_b_r,   i_b_i,
                      i_c_r,   i_c_i,
                      i_n_r,   i_n_i,
                      i_pos,i_neg,#v_og,
                      p_dc]

    # current injections ac side
    for ph in ['a','b','c','n']:
        i_s_r = sym.Symbol(f'i_vsc_{bus_ac_name}_{ph}_r', real=True)
        i_s_i = sym.Symbol(f'i_vsc_{bus_ac_name}_{ph}_i', real=True)  
        idx_r,idx_i = grid.node2idx(bus_ac_name,ph)
        grid.dae['g'] [idx_r] += -i_s_r
        grid.dae['g'] [idx_i] += -i_s_i
        i_s = i_s_r + 1j*i_s_i
        i_s_m = np.abs(i_s)
        grid.dae['h_dict'].update({f'i_vsc_{bus_ac_name}_{ph}_m':i_s_m})

    # current injections dc side
    idx_r,idx_i = grid.node2idx(f'{bus_dc_name}','a')
    grid.dae['g'] [idx_r] += i_pos 
    grid.dae['g'] [idx_i] += v_posi/1e3

    idx_r,idx_i = grid.node2idx(f'{bus_dc_name}','b')
    grid.dae['g'] [idx_r] += i_neg 
    grid.dae['g'] [idx_i] += v_negi/1e3  
    
      
    grid.dae['u_ini_dict'].update({f'v_dc_{bus_dc_name}_ref':800.0}) 
    grid.dae['u_ini_dict'].update({f'{str(p_a)}':0.0,f'{str(p_b)}':0.0,f'{str(p_c)}':0.0}) 
    grid.dae['u_ini_dict'].update({f'{str(q_a)}':0.0,f'{str(q_b)}':0.0,f'{str(q_c)}':0.0}) 

    grid.dae['u_run_dict'].update({f'v_dc_{bus_dc_name}_ref':800.0}) 
    grid.dae['u_run_dict'].update({f'{str(p_a)}':0.0,f'{str(p_b)}':0.0,f'{str(p_c)}':0.0}) 
    grid.dae['u_run_dict'].update({f'{str(q_a)}':0.0,f'{str(q_b)}':0.0,f'{str(q_c)}':0.0}) 


    #grid.dae['u'].pop(str(v_dc_a_r))
    grid.dae['params_dict'].update({f'A_loss_{bus_ac_name}':a_value,f'B_loss_{bus_ac_name}':b_value,f'C_loss_{bus_ac_name}':c_value})
    grid.dae['params_dict'].update({f'R_dc_{bus_dc_name}':1e-6})
    grid.dae['params_dict'].update({f'K_dc_{bus_dc_name}':1e-6})
    grid.dae['params_dict'].update({f'R_gdc_{bus_dc_name}':3.0})    

    grid.dae['xy_0_dict'].update({f'v_{bus_dc_name}_a_r':800.0,f'v_{bus_dc_name}_n_r':1.0})
    
    grid.dae['h_dict'].update({f'p_vsc_{bus_ac_name}':sym.re(s_a)+sym.re(s_b)+sym.re(s_c)+sym.re(s_n)})
    grid.dae['h_dict'].update({f'p_vsc_loss_{bus_ac_name}':(p_loss_total)})
    grid.dae['h_dict'].update({f'v_dc_{bus_dc_name}':v_dc})




# -*- coding: utf-8 -*-
"""
Created on Thu August 10 23:52:55 2022

@author: jmmauricio
"""

import numpy as np
import sympy as sym

def full_converter(grid,name,bus_name,data_dict):
    '''

    parameters
    ----------

    S_n: nominal power in VA
    F_n: nominal frequency in Hz
    X_v: coupling reactance in pu (base machine S_n)
    R_v: coupling resistance in pu (base machine S_n)
    K_delta: if K_delta>0.0 current generator is converted to reference machine 
    K_alpha: alpha gain to obtain Domega integral 

    inputs
    ------

    alpha: RoCoF in pu if K_alpha = 1.0
    omega_ref: frequency in pu
    v_ref: internal voltage reference

    example
    -------

    "vscs": [{"type":"vsc_l","S_n":1e6,"F_n":50.0,"X_s":0.1,"R_s":0.01}]

    v_t = m*v_dc/sqrt(6)

    v_t_pu*v_b_ac = m*v_dc_pu*v_b_dc/sqrt(6)

    v_dc_b = sqrt(2)*v_ac_b
    v_t_pu*v_b_ac = m*v_dc_pu*sqrt(2)*v_ac_b/sqrt(6)
    v_t_pu = m*v_dc_pu*sqrt(2)/sqrt(6) = m*v_dc_pu*sqrt(3)
    
    
    '''

    sin = sym.sin
    cos = sym.cos

    # inputs
    V_s = sym.Symbol(f"V_{bus_name}", real=True)
    theta_s = sym.Symbol(f"theta_{bus_name}", real=True)
    m = sym.Symbol(f"m_{name}", real=True)
    theta_t = sym.Symbol(f"theta_t_{name}", real=True)
    v_dc = sym.Symbol(f"v_dc_{name}", real=True)
    Dp_e_ref = sym.Symbol(f"Dp_e_ref_{name}", real=True)
    u_dummy = sym.Symbol(f"u_dummy_{name}", real=True)
      
    # dynamic states
    theta_tr = sym.Symbol(f"theta_tr_{name}", real=True)
    omega_t  = sym.Symbol(f"omega_t_{name}", real=True)
    omega_r  = sym.Symbol(f"omega_r_{name}", real=True)
    xi_beta  = sym.Symbol(f"xi_beta_{name}", real=True)
    beta     = sym.Symbol(f"beta_{name}", real=True)
    p_m_mppt_lpf = sym.Symbol(f"p_m_mppt_lpf_{name}", real=True)

    # algebraic states
    i_si = sym.Symbol(f"i_si_{name}", real=True)
    i_sr = sym.Symbol(f"i_sr_{name}", real=True)       
    p_s = sym.Symbol(f"p_s_{name}", real=True)
    q_s = sym.Symbol(f"q_s_{name}", real=True)
    p_s_ref = sym.Symbol(f"p_s_ref_{name}", real=True)
    nu_w = sym.Symbol(f"nu_w_{name}", real=True)
    
    # parameters
    S_n = sym.Symbol(f"S_n_{name}", real=True)
    F_n = sym.Symbol(f"F_n_{name}", real=True)            
    X_s = sym.Symbol(f"X_s_{name}", real=True)
    R_s = sym.Symbol(f"R_s_{name}", real=True)
    A_l = sym.Symbol(f"A_l_{name}", real=True)
    B_l = sym.Symbol(f"B_l_{name}", real=True)
    C_l = sym.Symbol(f"C_l_{name}", real=True)
    K_pow = sym.Symbol(f"K_pow_{name}", real=True)
    C_1,C_2,C_3 = sym.symbols(f"C_1_{name},C_2_{name},C_3_{name}", real=True)
    C_4,C_5,C_6 = sym.symbols(f"C_4_{name},C_5_{name},C_6_{name}", real=True)
    Nu_w_b,Lam_b = sym.symbols(f"Nu_w_b_{name},Lam_b_{name}", real=True)
    K_tr,D_tr,Omega_t_b = sym.symbols(f"K_tr_{name},D_tr_{name},Omega_t_b_{name}", real=True)
    C_p_b,T_beta,T_mppt,K_mppt = sym.symbols(f"C_p_b_{name},T_beta_{name},T_mppt_{name},K_mppt_{name}", real=True)
    K_p_beta = sym.Symbol(f"K_p_beta_{name}", real=True)
    K_i_beta = sym.Symbol(f"K_i_beta_{name}", real=True)
    H_t,H_r= sym.symbols(f"H_t_{name},H_r_{name}", real=True)
    K_tr,D_tr= sym.symbols(f"K_tr_{name},D_tr_{name}", real=True)
    params_list = ['S_n','F_n','X_s','R_s']
    
    # auxiliar
    ## VSC
    v_sr =  V_s*cos(theta_s)  # v_Q
    v_si =  V_s*sin(theta_s)  # v_D, e^(-j)
    Omega_b = 2*np.pi*F_n
    v_t_m = m*v_dc
    v_tr =  v_t_m*cos(theta_t)
    v_ti =  v_t_m*sin(theta_t)
    i_m = sym.sqrt(i_sr**2 + i_si**2)
    p_loss = A_l + B_l*i_m + C_l*i_m**2
    p_ac = p_s + R_s*i_m**2
    p_dc = p_ac + p_loss
    i_d = p_dc/v_dc

    ## Aerodinamic
    lam = Lam_b*(omega_t/Omega_t_b)/(nu_w/Nu_w_b) # pu
    inv_lam_i =  1/(lam + 0.08*beta) - 0.035/(beta**3 + 1.0)   
    c_p = C_1*(C_2*inv_lam_i - C_3*beta - C_4)*sym.exp(-C_5*inv_lam_i) + C_6*lam 
    c_p_pu = c_p/C_p_b # (pu)
    p_m_nosat = K_pow*c_p_pu*(nu_w/Nu_w_b)**3
    p_m = sym.Piecewise((0.0,p_m_nosat<0.0), (p_m_nosat,True))

    ## MPPT
    Lam_opt = Lam_b
    beta_mppt = 0.0
    nu_w_mppt = Nu_w_b*(omega_t/Omega_t_b)   
    inv_lam_i_mppt =  1/(Lam_opt) - 0.035   
    c_p_mppt = C_1*(C_2*inv_lam_i_mppt - C_4)*sym.exp(-C_5*inv_lam_i_mppt) + C_6*Lam_opt 
    c_p_mppt_pu = c_p_mppt/C_p_b # (pu)
    p_m_mppt_ref = K_pow*c_p_mppt_pu*(nu_w_mppt/Nu_w_b)**3
    omega_r_th = 0.2
    test_1 = (np.abs(Dp_e_ref)>0.01) & (omega_r>omega_r_th)
    p_m_mppt = sym.Piecewise((p_m_mppt_lpf,test_1), (p_m_mppt_ref,True))
    K_mppt = sym.Piecewise((0.0,test_1), (1.0,True))
    Dp_e = sym.Piecewise((Dp_e_ref,test_1), (Dp_e_ref * (1 + 50*(omega_r - omega_r_th)),True))
    p_s_mmpt_ref = p_m_mppt_lpf + Dp_e    

    ## Pitch
    Omega_r_b = Omega_t_b
    omega_r_max = Omega_r_b
    epsilon_beta_nosat = omega_r - omega_r_max
    epsilon_beta = sym.Piecewise((0.0,epsilon_beta_nosat<0),(epsilon_beta_nosat,True))
    beta_ref = K_p_beta*epsilon_beta + K_i_beta*xi_beta

    # dynamic equations            
    dtheta_tr = omega_t-omega_r - u_dummy
    domega_t  = 1.0/(2*H_t)*(p_m - K_tr*theta_tr - D_tr*(omega_t-omega_r))
    domega_r  = 1.0/(2*H_r)*(K_tr*theta_tr + D_tr*(omega_t-omega_r) - p_dc)
    dxi_beta  = epsilon_beta - xi_beta*1e-8
    dbeta     = 1.0/T_beta*(beta_ref - beta)
    dp_m_mppt_lpf  = K_mppt/T_mppt*(p_m_mppt_ref - p_m_mppt_lpf)

    # algebraic equations   
    g_i_si = v_ti - R_s*i_si - X_s*i_sr - v_si  
    g_i_sr = v_tr - R_s*i_sr + X_s*i_si - v_sr 
    g_p_s  =  i_sr*v_sr + i_si*v_si - p_s  
    g_q_s  =  i_sr*v_si - i_si*v_sr - q_s 
    g_p_s_ref = -p_s_ref + p_s_mmpt_ref


    # dae 
    grid.dae['f'] += [dtheta_tr,domega_t,domega_r,dxi_beta,dbeta,dp_m_mppt_lpf]
    grid.dae['x'] += [ theta_tr, omega_t, omega_r, xi_beta, beta, p_m_mppt_lpf]
    grid.dae['g'] += [g_i_si,g_i_sr,g_p_s,g_q_s,g_p_s_ref]
    grid.dae['y_ini'] += [  i_si,  i_sr,  p_s,  q_s, p_s_ref]  
    grid.dae['y_run'] += [  i_si,  i_sr,  p_s,  q_s, p_s_ref]  
    
    grid.dae['u_ini_dict'].update({f'{m}':1.0})
    grid.dae['u_run_dict'].update({f'{m}':1.0})

    grid.dae['u_ini_dict'].update({f'{theta_t}':0.0})
    grid.dae['u_run_dict'].update({f'{theta_t}':0.0})

    grid.dae['u_ini_dict'].update({f'{v_dc}':1.2})
    grid.dae['u_run_dict'].update({f'{v_dc}':1.2})

    grid.dae['u_ini_dict'].update({f'{Dp_e_ref}':0.0})
    grid.dae['u_run_dict'].update({f'{Dp_e_ref}':0.0})

    grid.dae['u_ini_dict'].update({f'{nu_w}':10.0})
    grid.dae['u_run_dict'].update({f'{nu_w}':10.0})

    grid.dae['u_ini_dict'].update({f'{u_dummy}':0.0})
    grid.dae['u_run_dict'].update({f'{u_dummy}':0.0})
 


    # outputs
    grid.dae['h_dict'].update({f"{str(p_s)}":p_s})
    grid.dae['h_dict'].update({f"{str(q_s)}":q_s}) 
    grid.dae['h_dict'].update({f"{str(i_si)}":i_si})
    grid.dae['h_dict'].update({f"{str(i_sr)}":i_sr})
    grid.dae['h_dict'].update({f"p_ac_{name}":p_ac})
    grid.dae['h_dict'].update({f"p_dc_{name}":p_dc})
    grid.dae['h_dict'].update({f"i_d_{name}":i_d})
    grid.dae['h_dict'].update({f"u_dummy_{name}":u_dummy})
    grid.dae['h_dict'].update({f"p_m_mppt_lpf_{name}":p_m_mppt_lpf})
    grid.dae['h_dict'].update({f"p_t_{name}":p_m})
    grid.dae['h_dict'].update({f"K_mppt_{name}":K_mppt})
    grid.dae['h_dict'].update({f"lam_{name}":lam})
    grid.dae['h_dict'].update({f"c_p_{name}":c_p})
    grid.dae['h_dict'].update({f"c_p_pu_{name}":c_p_pu})
    grid.dae['h_dict'].update({f"nu_w_mppt_{name}":nu_w_mppt})
    grid.dae['h_dict'].update({f"p_m_{name}":p_m})
    

    for item in params_list:       
        grid.dae['params_dict'].update({f"{item}_{name}":data_dict[item]}) 

    if 'A_l' in data_dict:
        grid.dae['params_dict'].update({f"A_l_{name}":data_dict['A_l']}) 
        grid.dae['params_dict'].update({f"B_l_{name}":data_dict['B_l']}) 
        grid.dae['params_dict'].update({f"C_l_{name}":data_dict['C_l']}) 
    else:
        grid.dae['params_dict'].update({f"A_l_{name}":0.005}) 
        grid.dae['params_dict'].update({f"B_l_{name}":0.005}) 
        grid.dae['params_dict'].update({f"C_l_{name}":0.005})         

    C_1 = 0.5176
    C_2 = 116
    C_3 = 0.4
    C_4 = 5.0
    C_5 = 21 
    C_6 = 0.0068
    Nu_w_b = 12
    Lam_b = 8.1
    Omega_t_b = 1.2
    K_pow = 0.73
    grid.dae['params_dict'].update({f"C_1_{name}":C_1})
    grid.dae['params_dict'].update({f"C_2_{name}":C_2})
    grid.dae['params_dict'].update({f"C_3_{name}":C_3})
    grid.dae['params_dict'].update({f"C_4_{name}":C_4})
    grid.dae['params_dict'].update({f"C_5_{name}":C_5})
    grid.dae['params_dict'].update({f"C_6_{name}":C_6})
    grid.dae['params_dict'].update({f"Nu_w_b_{name}":Nu_w_b})
    grid.dae['params_dict'].update({f"Lam_b_{name}":Lam_b})
    grid.dae['params_dict'].update({f"Omega_t_b_{name}":Omega_t_b})
    grid.dae['params_dict'].update({f"K_pow_{name}":K_pow})
    ## 2 mass mechanical model
    omega_ref = 2*np.pi*0.5
    zeta_ref = 0.05
    H_t = 3.0
    H_r = 1.0
    grid.dae['params_dict'].update({f"H_t_{name}":H_t})
    grid.dae['params_dict'].update({f"H_r_{name}":H_r})
    K_tr = 2*H_r*H_t*omega_ref**2/(H_r + H_t)
    D_tr =  4*np.pi*H_r*H_t*zeta_ref*omega_ref/(H_r + H_t)
    grid.dae['params_dict'].update({f"K_tr_{name}":K_tr})
    grid.dae['params_dict'].update({f"D_tr_{name}":D_tr})
    ## c_p_b:
    beta_b = 0.0
    lam = Lam_b
    inv_lam_i =  1/(lam + 0.08*beta_b) - 0.035/(beta_b**3 + 1.0)   
    C_p_b = C_1*(C_2*inv_lam_i - C_3*beta_b - C_4)*np.exp(-C_5*inv_lam_i) + C_6*lam 
    grid.dae['params_dict'].update({f"C_p_b_{name}":C_p_b})

    grid.dae['params_dict'].update({f"T_beta_{name}":2.0})
    grid.dae['params_dict'].update({f"K_p_beta_{name}":100.0})
    grid.dae['params_dict'].update({f"K_i_beta_{name}":1.0})
    grid.dae['params_dict'].update({f"T_mppt_{name}":5})

    
    # initial conditions
    beta_0 = 0.0  # initial input
    p_ref = 0.5*0.657/0.9 # initial input
    inv_lam_i =  1/(lam + 0.08*beta_0) - 0.035/(beta_0**3 + 1.0)   
    c_p = C_1*(C_2*inv_lam_i - C_3*beta_0 - C_4)*np.exp(-C_5*inv_lam_i) + C_6*lam 
    c_p_pu = c_p/C_p_b # (pu)
    #p_m_ref = K_p*c_p*(nu_w/nu_w_b)**3
    nu_w_0 = (p_ref/(K_pow*c_p_pu))**(1.0/3.0)*Nu_w_b
    omega_t_0 = lam*(nu_w_0/Nu_w_b)/Lam_b*Omega_t_b

    grid.dae['xy_0_dict'].update({str(p_s):p_ref})
    grid.dae['xy_0_dict'].update({str(theta_tr):p_ref/K_tr})
    grid.dae['xy_0_dict'].update({str(omega_t):omega_t_0})
    grid.dae['xy_0_dict'].update({str(omega_r):omega_t_0})
    grid.dae['xy_0_dict'].update({str(beta):beta_0})
    grid.dae['xy_0_dict'].update({str(p_s_ref):p_ref})
    grid.dae['xy_0_dict'].update({str(nu_w):nu_w_0})
    grid.dae['xy_0_dict'].update({str(i_sr):0.1})

    p_W   = p_s * S_n
    q_var = q_s * S_n

    return p_W,q_var





# -*- coding: utf-8 -*-
"""
Created on Thu August 10 23:52:55 2022

@author: jmmauricio
"""

import numpy as np
import sympy as sym
from pydae.urisi.vscs.ac3ph4wgfpi2 import ac3ph4wgfpi2
from pydae.urisi.genapes.ac3ph4w_ideal import ac3ph4w_ideal
from pydae.urisi.genapes.ac3ph3w_ideal import ac3ph3w_ideal
from pydae.urisi.genapes.dc_ideal import dc_ideal

def add_genapes(grid):

    buses = grid.data['buses']
    buses_list = [bus['name'] for bus in buses]

    for item in grid.data['genapes']:
        if item['type'] == 'ac3ph4w_ideal':
            ac3ph4w_ideal(grid,item)
        if item['type'] == 'ac3ph3w_ideal':
            ac3ph3w_ideal(grid,item)       
        if item['type'] == 'dc_ideal':
            dc_ideal(grid,item)       


import numpy as np
import matplotlib.pyplot as plt
import ipywidgets
plt.style.use('https://raw.githubusercontent.com/pydae/pydae/master/src/pydae/edashboards/presentation.mplstyle')
colors = plt.rcParams['axes.prop_cycle'].by_key()['color']

class dashboard():
    
    def __init__(self, model):
        
        model.Dt = 0.01
        model.decimation = 1
        model.ini({'v_ref_1':1.0,'P_1':0.0},'xy_0.json')
        model.run(20.0,{})
        model.post();
        
        self.model = model
        
        self.colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
        self.widgets()
        
    def widgets(self):
        
        model = self.model
        colors = self.colors

        plt.ioff()
        plt.clf()


        fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(9, 4), frameon=False)
        fig.canvas.toolbar_visible = False

        self.line_v_f = axes[0,0].plot(model.Time, model.get_values('v_f_1'), label='$\sf v_f$', color=colors[3])
        self.line_p_m = axes[0,0].plot(model.Time, model.get_values('p_m_1'), label='$\sf p_m$', color=colors[4])

        self.line_omega = axes[1,0].plot(model.Time, model.get_values('omega_1'), label='$\sf \omega$', color=colors[1])
        self.line_v_1 = axes[0,1].plot(model.Time, model.get_values('V_1'), label='$\sf V_1$', color=colors[5])
        #line_theta_1 = axes[0,1].plot(T, Y[:,syst.y_list.index('theta_1')], label='$\sf \\theta_1$')
        self.line_p_t = axes[1,1].plot(model.Time, model.get_values('p_g_1'), label='$\sf P_t$', color=colors[2])
        self.line_q_t = axes[1,1].plot(model.Time, model.get_values('q_g_1'), label='$\sf Q_t$', color=colors[0])

        y_labels = ['$\delta$','$\omega$','$P_t$']

        axes[0,0].set_ylim((0,3))
        axes[1,0].set_ylim((0.95,1.05))
        axes[0,1].set_ylim((0.8,1.2))
        axes[1,1].set_ylim((-0.5,1.5))

        axes[0,0].grid(True)
        axes[1,0].grid(True)
        axes[0,1].grid(True)
        axes[1,1].grid(True)
        axes[0,0].legend(loc='best')
        axes[1,0].legend(loc='best')
        axes[0,1].legend(loc='best')
        axes[1,1].legend(loc='best')

        axes[1,0].set_xlabel('Time (s)')  
        axes[1,1].set_xlabel('Time (s)') 

        fig.tight_layout()
        
        self.fig = fig
        #axes[0].set_title('Par en función de la velocidad')
        #axes[1].set_title('Corriente en función de la velocidad')


        self.sld_P_l = ipywidgets.FloatSlider(orientation='horizontal',description = "P<sub>l</sub>", 
                                        value=model.get_value('P_2'), min=0.0,max= 120, 
                                        step=1)
        self.sld_Q_l = ipywidgets.FloatSlider(orientation='horizontal',description = "Q<sub>l</sub>", 
                                        value=model.get_value('Q_2'), min=-100.0,max= 100, 
                                        step=1)

        self.sld_v_ref_1 = ipywidgets.FloatSlider(orientation='horizontal',description = "V<sup>*</sup>", 
                                        value=model.get_value('v_ref_1'), min=0.85,max= 1.15, 
                                        step=.01)

        self.prog_c = ipywidgets.IntProgress(
            value=100,
            min=0,
            max=120,
            step=1,
            description='SM Load:',
            bar_style='', # 'success', 'info', 'warning', 'danger' or ''
            orientation='horizontal' 
        )

        self.prog_damp = ipywidgets.IntProgress(
            value=10,
            min=0,
            max=20,
            step=1,
            description='ζ = 1.0',
            bar_style='', # 'success', 'info', 'warning', 'danger' or ''
            orientation='horizontal' 
        )
        
    def update(self,change):
        
        model = self.model

        P_l = self.sld_P_l.value
        Q_l = self.sld_Q_l.value
        v_ref_1 = self.sld_v_ref_1.value


        #model = smib.smib_class()
        #model.Dt = 0.01
        #model.decimation = 1
        model.Dt = 0.01
        model.ini({'P_1':0.0,'Q_1':0.0,'v_ref_1':1.0},'xy_0.json')
        model.run( 1.0,{})
        model.run(10,{'P_1':-P_l*1e6,'Q_1':-Q_l*1e6,'v_ref_1':v_ref_1})
        model.Dt = 0.1
        model.run(20,{'P_1':-P_l*1e6,'Q_1':-Q_l*1e6,'v_ref_1':v_ref_1})

        model.post();


        self.line_v_f[0].set_data(model.Time, model.get_values('v_f_1'))
        self.line_p_m[0].set_data(model.Time, model.get_values('p_m_1'))

        self.line_omega[0].set_data(model.Time, model.get_values('omega_1'))
        self.line_v_1[0].set_data(model.Time, model.get_values('V_1'))
        #line_theta_1 = axes[0,1].plot(T, Y[:,syst.y_list.index('theta_1')], label='$\sf \\theta_1$')
        self.line_p_t[0].set_data(model.Time, model.get_values('p_g_1'))
        self.line_q_t[0].set_data(model.Time, model.get_values('q_g_1'))

        i_d, i_q = model.get_mvalue(['i_d_1','i_q_1'])
        c = (i_d**2+i_q**2)*0.5

        self.prog_c.bar_style = 'success'
        if c>0.9:
            self.prog_c.bar_style = 'warning'
        if c>1.0:
            self.prog_c.bar_style = 'danger'
        self.prog_c.value = 100*c

        self.fig.canvas.draw_idle()


    def show(self):

        self.sld_P_l.observe(self.update, names='value')
        self.sld_Q_l.observe(self.update, names='value')

        self.sld_v_ref_1.observe(self.update, names='value')

        layout_row1 = ipywidgets.HBox([self.fig.canvas])
        PQ = ipywidgets.VBox([self.sld_P_l,self.sld_Q_l])
        layout_row2 = ipywidgets.HBox([PQ,self.sld_v_ref_1,self.prog_c])

        layout = ipywidgets.VBox([layout_row1,layout_row2])
        self.layout = layout
        display(self.layout)
     


# -*- coding: utf-8 -*-
"""
Created on Thu August 10 23:52:55 2022

@author: jmmauricio
"""

import numpy as np
import sympy as sym

def uvsg(grid,name,bus_name,data_dict):
    """
    # auxiliar equations
    Omega_b = 2*np.pi*F_n
    omega_s = omega_coi
    v_D = V*sin(theta)  # e^(-j)
    v_Q = V*cos(theta) 
    v_d = v_D * cos(delta) - v_Q * sin(delta)   
    v_q = v_D * sin(delta) + v_Q * cos(delta)

    Domega = x_v + K_p * (p_ref - p)
    e_dv = 0.0
    epsilon_v = v_ref - V
    i_d = i_d_ref
    i_q = i_q_ref
    omega_v = Domega + 1.0
    q_ref_0 = K_p_v * epsilon_v + K_i_v * xi_v

    # dynamical equations
    ddelta   = Omega_b*(omega_v - omega_s) - K_delta*delta
    dx_v = K_i*(p_ref - p) - K_g*(omega_v - 1.0)
    de_qm = 1.0/T_q * (q - q_ref_0 - q_ref - e_qm) 
    dxi_v = epsilon_v # PI agregado

    # algebraic equations
    g_i_d_ref  = e_dv - R_v * i_d_ref - X_v * i_q_ref - v_d 
    g_i_q_ref  = e_qv - R_v * i_q_ref + X_v * i_d_ref - v_q 
    g_p  = v_d*i_d + v_q*i_q - p  
    g_q  = v_d*i_q - v_q*i_d - q 
    g_e_qv = 1.0 - e_qv - K_q*e_qm 

    {"bus":"1","type":"vsg_ll",'S_n':10e6,'F_n':50,'K_delta':0.0,
    'R_v':0.01,'X_v':0.1,'K_p':1.0,'K_i':0.1,'K_g':0.0,'K_q':20.0,
    'T_q':0.1,'K_p_v':1e-6,'K_i_v':1e-6}
    
    """

    sin = sym.sin
    cos = sym.cos

    ctrl_data = data_dict['ctrl']

    buses = grid.data['buses']
    buses_list = [bus['name'] for bus in buses]                  

    idx_bus = buses_list.index(bus_name) # get the number of the bus where the syn is connected
    if not 'idx_powers' in buses[idx_bus]: buses[idx_bus].update({'idx_powers':0})
    buses[idx_bus]['idx_powers'] += 1

    # inputs:
    q_l,q_r = sym.symbols(f'q_l_{name},q_r_{name}', real=True)
    p_l,p_r = sym.symbols(f'p_l_{name},p_r_{name}', real=True)
    v_dc = sym.Symbol(f"v_dc_{name}", real=True)   
    Ddelta_ff,Domega_ff = sym.symbols(f'Ddelta_ff_{name},Domega_ff_{name}', real=True)
    v_u = sym.Symbol(f"v_u_{name}", real=True)  
    v_u_ref  = sym.Symbol(f"v_u_ref_{name}", real=True) 
    i_sact_ref = sym.Symbol(f"i_sact_ref_{name}", real=True) 
    i_srea_ref = sym.Symbol(f"i_srea_ref_{name}", real=True)

    # dynamical states:
    delta = sym.Symbol(f"delta_{name}", real=True)   
    xi_p  = sym.Symbol(f"xi_p_{name}", real=True)   
    xi_q  = sym.Symbol(f"xi_q_{name}", real=True)   
    xi_u  = sym.Symbol(f"xi_u{name}", real=True)   

    # algebraic states:
    p_s = sym.Symbol(f'p_s_{name}', real=True)
    q_s = sym.Symbol(f'q_s_{name}', real=True)
    e_qv = sym.Symbol(f"e_qv_{name}", real=True)
    v_td_ref = sym.Symbol(f"v_td_ref_{name}", real=True)
    v_tq_ref = sym.Symbol(f"v_tq_ref_{name}", real=True) 
    i_sd_mid = sym.Symbol(f"i_sd_mid_{name}", real=True)
    i_sq_mid = sym.Symbol(f"i_sq_mid_{name}", real=True) 
    i_sd_fault = sym.Symbol(f"i_sd_fault_{name}", real=True) 
    i_sq_fault = sym.Symbol(f"i_sq_fault_{name}", real=True)
    omega_v  = sym.Symbol(f"omega_v_{name}", real=True) 
    p_u = sym.Symbol(f"p_u_{name}", real=True) 

    # params:
    S_n,F_n = sym.symbols(f'S_n_{name},F_n_{name}', real=True)
    R_v,X_v = sym.symbols(f'R_v_{name},X_v_{name}', real=True)
    K_p,T_p = sym.symbols(f'K_p_{name},T_p_{name}', real=True)
    K_q,T_q = sym.symbols(f'K_q_{name},T_q_{name}', real=True)
    K_u,T_u = sym.symbols(f'K_u_{name},T_u_{name}', real=True)
    R_s,X_s = sym.symbols(f'R_s_{name},X_s_{name}', real=True)
    K_u_max,K_ui,DV_th = sym.symbols(f'K_u_max_{name},K_ui_{name},DV_th_{name}', real=True)
    params_list = ['F_n','R_v','X_v','K_p','T_p','K_q','T_q','K_u','T_u']

    # auxiliar variables and constants
    omega_coi = sym.Symbol("omega_coi", real=True) # from global system
    V_s = sym.Symbol(f"V_{bus_name}", real=True) # from global system
    theta_s = sym.Symbol(f"theta_{bus_name}", real=True) # from global system 
    m = sym.Symbol(f"m_{name}", real=True) 
    theta_t = sym.Symbol(f"theta_t_{name}", real=True) 

    # auxiliar equations
    Omega_b = 2*np.pi*F_n
    omega_s = omega_coi
    e_dv = 0
    delta_ff =  delta + Ddelta_ff

    park = 'easy-res'
    # easy-ref park
    if park == 'easy-res': 

        v_sr = V_s*cos(theta_s)  # v_sr
        v_si = V_s*sin(theta_s)  # v_si
        v_salpha =-v_si
        v_sbeta  =-v_sr

        v_sd = v_salpha * cos(delta_ff) + v_sbeta * sin(delta_ff)   
        v_sq =-v_salpha * sin(delta_ff) + v_sbeta * cos(delta_ff)

        v_talpha_ref =  v_td_ref * cos(delta_ff) - v_tq_ref * sin(delta_ff)   
        v_tbeta_ref  =  v_td_ref * sin(delta_ff) + v_tq_ref * cos(delta_ff) 

        v_ti_ref = -v_talpha_ref
        v_tr_ref = -v_tbeta_ref   

    v_m = (v_sd**2 + v_sq**2 + 0.0001)**0.5
    lvrt = sym.Piecewise((1.0,v_m<(0.8-DV_th)),(1.0*(0.8-v_m)/DV_th,v_m<0.8),(0.0,True))
    p_ref = p_l + p_r + p_u
    q_ref = q_l + q_r
    epsilon_p = (p_ref - p_s)*(1-lvrt)
    epsilon_q = q_ref - q_s
    epsilon_u = v_u_ref - v_u
    p_u_l_ref = -K_u*epsilon_u
    
    p_u_ref = sym.Piecewise((K_u_max*(v_u-0.1)+p_u_l_ref,v_u<0.1),
                            (K_u_max*(v_u-0.9)+p_u_l_ref,v_u>0.9),
                            (p_u_l_ref,True))


    #i_sd_ref = sym.Piecewise((i_sd_fault,lvrt>0.5),(i_sd_mid,True))
    #i_sq_ref = sym.Piecewise((i_sq_fault,lvrt>0.5),(i_sq_mid,True))
    i_sd_ref = (lvrt)*i_sd_fault + (lvrt-1)*i_sd_mid
    i_sq_ref = (lvrt)*i_sq_fault + (lvrt-1)*i_sq_mid

    # from controller to vsc hardware
    m_ref = sym.sqrt(v_tr_ref**2 + v_ti_ref**2)/v_dc
    theta_t_ref = sym.atan2(v_ti_ref,v_tr_ref) 


    # dynamic equations            
    ddelta = Omega_b*(omega_v - omega_s)
    dxi_p = epsilon_p
    dxi_q = epsilon_q
    dxi_u = epsilon_u



    # algebraic equations
    g_v_td_ref  = v_td_ref - R_s * i_sd_ref - X_s * i_sq_ref - v_sd   # ctrl1 i_sd control
    g_v_tq_ref  = v_tq_ref - R_s * i_sq_ref + X_s * i_sd_ref - v_sq   # ctrl1 i_sq control
    g_i_sd_mid  = e_dv - R_v * i_sd_mid - X_v * i_sq_mid - v_sd   # ctrl3 virtual impedance 
    g_i_sq_mid  = e_qv - R_v * i_sq_mid + X_v * i_sd_mid - v_sq   # ctrl3 virtual impedance
    g_i_sd_fault  = (i_sd_fault*v_sd + i_sq_fault*v_sq)/v_m - i_sact_ref  
    g_i_sq_fault  = (i_sq_fault*v_sd - i_sd_fault*v_sq)/v_m - i_srea_ref
    g_omega_v = -omega_v + K_p*(epsilon_p + xi_p/T_p)    # p PI -> omega_v
    g_e_qv = -e_qv   + K_q*(epsilon_q + xi_q/T_q)        # q PI -> e_qv
    g_p_u  = -p_u + p_u_ref - K_ui*xi_u
    g_m  = m-m_ref   # to the VSC
    g_theta_t  = theta_t-theta_t_ref # to the VSC

    # DAE system update
    grid.dae['f'] += [ddelta,dxi_p,dxi_q,dxi_u]
    grid.dae['x'] += [ delta, xi_p, xi_q, xi_u]
    grid.dae['g'] +=     [g_v_td_ref,g_v_tq_ref,g_i_sd_mid,g_i_sq_mid,g_i_sd_fault,g_i_sq_fault,g_omega_v,g_e_qv,g_p_u,g_m,g_theta_t]
    grid.dae['y_ini'] += [  v_td_ref,  v_tq_ref,  i_sd_mid,  i_sq_mid,  i_sd_fault,  i_sq_fault,  omega_v,  e_qv,  p_u,  m,  theta_t]
    grid.dae['y_run'] += [  v_td_ref,  v_tq_ref,  i_sd_mid,  i_sq_mid,  i_sd_fault,  i_sq_fault,  omega_v,  e_qv,  p_u,  m,  theta_t]
            
    # default inputs
    grid.dae['u_ini_dict'].update({f'p_l_{name}':0.0})
    grid.dae['u_ini_dict'].update({f'q_l_{name}':0.0})
    grid.dae['u_ini_dict'].update({f'p_r_{name}':0.0})
    grid.dae['u_ini_dict'].update({f'q_r_{name}':0.0})
    grid.dae['u_ini_dict'].update({f'Ddelta_ff_{name}':0.0})
    grid.dae['u_ini_dict'].update({f'Domega_ff_{name}':0.0})
    grid.dae['u_ini_dict'].update({f'v_u_ref_{name}':0.5})
    grid.dae['u_ini_dict'].update({f'i_sact_ref_{name}':0.0})
    grid.dae['u_ini_dict'].update({f'i_srea_ref_{name}':1.0})

    grid.dae['u_run_dict'].update({f'p_l_{name}':0.0})
    grid.dae['u_run_dict'].update({f'q_l_{name}':0.0})
    grid.dae['u_run_dict'].update({f'p_r_{name}':0.0})
    grid.dae['u_run_dict'].update({f'q_r_{name}':0.0})
    grid.dae['u_run_dict'].update({f'Ddelta_ff_{name}':0.0})
    grid.dae['u_run_dict'].update({f'Domega_ff_{name}':0.0})
    grid.dae['u_run_dict'].update({f'v_u_ref_{name}':0.5})
    grid.dae['u_run_dict'].update({f'i_sact_ref_{name}':0.0})
    grid.dae['u_run_dict'].update({f'i_srea_ref_{name}':1.0})

    # default parameters
    for item in params_list:
        grid.dae['params_dict'].update({item + f'_{name}':ctrl_data[item]})
    grid.dae['params_dict'].update({f'K_u_max_{name}':2.0})
    grid.dae['params_dict'].update({f'K_ui_{name}':0.001})
    grid.dae['params_dict'].update({f'DV_th_{name}':0.1})


    # add speed*H term to global for COI speed computing
    H = 4.0
    grid.H_total += H
    grid.omega_coi_numerator += omega_v*H*S_n
    grid.omega_coi_denominator += H*S_n

    # DAE outputs update
    grid.dae['h_dict'].update({f"omega_v_{name}":omega_v})
    grid.dae['h_dict'].update({f"p_ref_{name}":p_ref})
    grid.dae['h_dict'].update({f"q_ref_{name}":q_ref})
    grid.dae['h_dict'].update({f"i_sd_ref_{name}":i_sd_ref})
    grid.dae['h_dict'].update({f"i_sq_ref_{name}":i_sq_ref})
    grid.dae['h_dict'].update({f"delta_ff_{name}":delta_ff})
    grid.dae['h_dict'].update({f"lvrt_{name}":lvrt})
    

    grid.dae['xy_0_dict'].update({str(e_qv):1.0}) 
    grid.dae['xy_0_dict'].update({str(v_tq_ref):1.0}) 

    p_W   = p_s * S_n
    q_var = q_s * S_n

    return p_W,q_var


if __name__ == "__main__":
    pass  


import pydae.build_cffi as db
from pydae.bmapu import bmapu_builder
from pydae.build_v2 import builder

import matplotlib.pyplot as plt
plt.style.use(r'https://raw.githubusercontent.com/pydae/pydae/master/src/pydae/edashboards/presentation.mplstyle')

import numpy as np
import ipywidgets
import pydae.svg_tools as st 
import json
 


class dashboard():
    
    def __init__(self):

        pass


    def ini(self):
    
        import spvib_dq_d

        model = spvib_dq_d.model()
        params = {}
        params.update({f'irrad_POI':1000+(np.random.rand()-0.5)*0})
        params.update({f'p_s_ppc_POI':0.5,f'q_s_ppc_POI':0})
        params.update({f'N_pv_s_POI':20, f'N_pv_p_POI':200})

        #model.report_u()
        #model.report_x()
        #model.report_y()
        model.ini(params,'xy_0.json')

        model.run(3.0,{})
        model.post();
        
        self.model = model
        
        #plt.style.use('presentation.mplstyle')
        self.colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
        #self.widgets()
        
    def widgets(self):
        
        model = self.model
        colors = self.colors

        plt.ioff()
        plt.clf()


        fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(9, 4), frameon=False)
        fig.canvas.toolbar_visible = False

        self.line_p = axes[0].plot(model.Time, model.get_values('p_s_POI'), label='$\sf p$', color=colors[0])
        self.line_q = axes[1].plot(model.Time, model.get_values('q_s_POI'), label='$\sf q$', color=colors[1])

        # self.line_omega = axes[1,0].plot(model.Time, model.get_values('omega_1'), label='$\sf \omega$', color=colors[1])
        # self.line_v_1 = axes[0,1].plot(model.Time, model.get_values('V_1'), label='$\sf V_1$', color=colors[5])
        # #line_theta_1 = axes[0,1].plot(T, Y[:,syst.y_list.index('theta_1')], label='$\sf \\theta_1$')
        # self.line_p_t = axes[1,1].plot(model.Time, model.get_values('p_g_1'), label='$\sf P_t$', color=colors[2])
        # self.line_q_t = axes[1,1].plot(model.Time, model.get_values('q_g_1'), label='$\sf Q_t$', color=colors[0])

        # y_labels = ['$\delta$','$\omega$','$P_t$']

        axes[0].set_ylim((0,1.2))
        axes[1].set_ylim((-1.2,1.2))
        # axes[0].set_ylim((0.8,1.2))
        # axes[1].set_ylim((-0.5,1.5))

        # axes[0,0].grid(True)
        # axes[1,0].grid(True)
        # axes[0,1].grid(True)
        # axes[1,1].grid(True)
        # axes[0,0].legend(loc='best')
        # axes[1,0].legend(loc='best')
        # axes[0,1].legend(loc='best')
        # axes[1,1].legend(loc='best')

        axes[0].set_xlabel('Time (s)')  
        axes[1].set_xlabel('Time (s)') 

        fig.tight_layout()
        fig.savefig('results.svg')
        
        self.fig = fig
        axes[0].set_title('Active Power (pu)')
        axes[1].set_title('Reactive Power (pu)')

        self.sld_p_ppc = ipywidgets.FloatSlider(orientation='horizontal',description = "p<sub>ppc</sub>", 
                                        value=model.get_value('p_s_ppc_POI'), min=0.0,max= 1.5, 
                                        step=.1)
        self.sld_q_ppc = ipywidgets.FloatSlider(orientation='horizontal',description = "q<sub>ppc</sub>", 
                                        value=model.get_value('q_s_ppc_POI'), min=-1.2,max= 1.2, 
                                        step=.1)

        self.sld_T_lp1p = ipywidgets.FloatSlider(orientation='horizontal',description = "T_lp1p", 
                                        value=model.get_value('T_lp1p_POI'), min=0.01,max= 2.0, 
                                        step=.01)
        self.sld_T_lp2p = ipywidgets.FloatSlider(orientation='horizontal',description = "T_lp2p", 
                                        value=model.get_value('T_lp2p_POI'), min=0.01,max= 2.0, 
                                        step=.01)
        self.sld_T_lp1q = ipywidgets.FloatSlider(orientation='horizontal',description = "T_lp1q", 
                                        value=model.get_value('T_lp1q_POI'), min=0.01,max= 2.0, 
                                        step=.01)
        self.sld_T_lp2q = ipywidgets.FloatSlider(orientation='horizontal',description = "T_lp2q", 
                                        value=model.get_value('T_lp2q_POI'), min=0.01,max= 2.0, 
                                        step=.01)
        self.sld_PRampUp = ipywidgets.FloatSlider(orientation='horizontal',description = "PRampUp", 
                                        value=model.get_value('PRampUp_POI'), min=0.1,max= 4.0, 
                                        step=.1)
        self.sld_PRampDown = ipywidgets.FloatSlider(orientation='horizontal',description = "PRampDown", 
                                        value=model.get_value('PRampDown_POI'), min=-4.0,max= -0.1, 
                                        step=.1)

        self.sld_QRampUp = ipywidgets.FloatSlider(orientation='horizontal',description = "QRampUp", 
                                        value=model.get_value('QRampUp_POI'), min=0.1,max= 4.0, 
                                        step=.1)
        self.sld_QRampDown = ipywidgets.FloatSlider(orientation='horizontal',description = "QRampDown", 
                                        value=model.get_value('QRampDown_POI'), min=-4.0,max= -0.1, 
                                        step=.1)
        
        self.s = st.svg(r"results.svg")

        self.html = ipywidgets.HTML(
            value= self.s.tostring(),
            placeholder='',
            description='',
        )

        self.btn_export = ipywidgets.Button(
            description='Export',
            disabled=False,
            button_style='', # 'success', 'info', 'warning', 'danger' or ''
            tooltip='Export parameters',
            icon='check')
        
        self.name = ipywidgets.Text(
            value='POI',
            placeholder='id name',
            description='Name:',
            disabled=False
        )
        
        
        self.export_text = ipywidgets.Textarea(
            value=' ',
            placeholder=' ',
            description='Params.:',
            disabled=False)
        
        tab_refs = ipywidgets.HBox([ipywidgets.HBox([self.sld_p_ppc,self.sld_q_ppc])])

        column_1 = ipywidgets.VBox([self.sld_T_lp1p,self.sld_T_lp2p,self.sld_PRampDown,self.sld_PRampUp])
        column_2 = ipywidgets.VBox([self.sld_T_lp1q,self.sld_T_lp2q,self.sld_QRampDown,self.sld_QRampUp])
        column_3 = ipywidgets.VBox([self.btn_export,self.name,self.export_text]) 
        tab_dynamics = ipywidgets.HBox([column_1,column_2,column_3])

        self.tabs = ipywidgets.Tab()
        self.tabs.children = [tab_refs,tab_dynamics]

        self.tabs.set_title(0, 'References')
        self.tabs.set_title(1, 'VSC Dynamics')
        
    def update(self,change):
        
        model = self.model

        p_ppc = self.sld_p_ppc.value
        q_ppc = self.sld_q_ppc.value

        T_lp1p = self.sld_T_lp1p.value
        T_lp2p = self.sld_T_lp2p.value
        T_lp1q = self.sld_T_lp1q.value
        T_lp2q = self.sld_T_lp2q.value

        PRampUp = self.sld_PRampUp.value
        QRampUp = self.sld_QRampUp.value
        PRampDown = self.sld_PRampDown.value
        QRampDown = self.sld_QRampDown.value

        model.decimation = 1
        model.Dt = 0.01
        model.ini({'p_s_ppc_POI':0.5,'q_s_ppc_POI':0.0,
                   'T_lp1p_POI':T_lp1p,'T_lp2p_POI':T_lp2p,
                   'T_lp1q_POI':T_lp1q,'T_lp2q_POI':T_lp2q,
                   'PRampUp_POI':PRampUp,'QRampUp_POI':QRampUp,
                   'PRampDown_POI':PRampDown,'QRampDown_POI':QRampDown},'xy_0.json')
        model.run( 1.0,{})
        model.run(3,{'p_s_ppc_POI':p_ppc,'q_s_ppc_POI':q_ppc})

        model.post();

        self.line_p[0].set_data(model.Time, model.get_values('p_s_POI'))
        self.line_q[0].set_data(model.Time, model.get_values('q_s_POI'))

        self.fig.canvas.draw_idle()
        self.fig.savefig('results.svg')

        self.s = st.svg(r"results.svg")

        self.html.value = self.s.tostring()

    def export_params(self,change):
        params_list = ['T_lp1p','T_lp2p','PRampDown','PRampUp','T_lp1q','T_lp2q','QRampDown','QRampUp']
        name = 'POI'
        name_toexport = self.name.value
        params_dict = {}
        for item in params_list:
            params_dict.update({f'{item}_{name_toexport}':self.model.get_value(f'{item}_{name}')})

        self.export_text.value = json.dumps(params_dict)


                

    def show(self):

        self.sld_p_ppc.observe(self.update, names='value')
        self.sld_q_ppc.observe(self.update, names='value')
        self.sld_T_lp1p.observe(self.update, names='value')
        self.sld_T_lp2p.observe(self.update, names='value')
        self.sld_T_lp1q.observe(self.update, names='value')
        self.sld_T_lp2q.observe(self.update, names='value')
        self.sld_PRampUp.observe(self.update, names='value')
        self.sld_QRampUp.observe(self.update, names='value')
        self.sld_PRampDown.observe(self.update, names='value')
        self.sld_QRampDown.observe(self.update, names='value')
        self.btn_export.on_click(self.export_params)   
        layout_row1 = ipywidgets.HBox([self.html])

        layout = ipywidgets.VBox([layout_row1,
                                  self.tabs])
        self.layout = layout
        display(self.layout)

    def build(self):

        S_pv_mva = 1.0

        data = {
            "system":{"name":f"spvib_dq_d","S_base":100e6,"K_p_agc":0.0,"K_i_agc":0.0,"K_xif":0.01},
            "buses":[
                {"name": "POI","P_W":0.0,"Q_var":0.0,"U_kV":0.4},
                {"name":"GRID","P_W":0.0,"Q_var":0.0,"U_kV":0.4}
            ],
            "lines":[
                {"bus_j":"POI","bus_k": "GRID","X_pu":0.01,"R_pu":0.0,"Bs_pu":0.0,"S_mva":1}
                ],
            "pvs":[{"bus":"POI","type":"pv_dq_d","S_n":1e6,"U_n":400.0,"F_n":50.0,"X_s":0.1,"R_s":0.01,"monitor":False,
                                    "I_sc":8,"V_oc":42.1,"I_mp":3.56,"V_mp":33.7,"K_vt":-0.160,"K_it":0.065,"N_pv_s":25,"N_pv_p":250}],
            "sources":[{"type":"genape",
                "bus":"GRID","S_n":1000e6,"F_n":50.0,"X_v":0.001,"R_v":0.0,
                "K_delta":0.001,"K_alpha":1e-6}]
            }

        grid = bmapu_builder.bmapu(data)

        grid.uz_jacs = False
        grid.verbose = False
        grid.construct(f'spvib_dq_d')
        b = builder(grid.sys_dict,verbose=True)
        b.dict2system()
        b.functions()
        b.jacobians()
        b.cwrite()
        b.template()
        b.compile()









# -*- coding: utf-8 -*-
"""
Created on Thu August 10 23:52:55 2022

@author: jmmauricio
"""

import numpy as np
import sympy as sym

def vsc_acgf_dcp(grid,name,bus_dc,bus_ac,data_dict):
    """
    # auxiliar
    
    """

    # inputs from grid
    V_s_m   = sym.Symbol(f"V_{bus_ac}", real=True)
    theta_s = sym.Symbol(f"theta_{bus_ac}", real=True)

    # inputs
    V_t_m = sym.Symbol(f"V_t_m_{bus_ac}", real=True)
    theta_t = sym.Symbol(f"theta_t_{bus_ac}", real=True)
 
    # algebraic states
    i_s_r,i_s_i = sym.symbols(f"i_s_r_{bus_ac},i_s_i_{bus_ac}", real=True)


    # parameters
    S_n = sym.Symbol(f"S_n_{name}", real=True)
    R_s = sym.Symbol(f"R_s_{name}", real=True)
    X_s = sym.Symbol(f"X_s_{name}", real=True)
    Losses_s,Losses_p = sym.symbols(f"Losses_s_{name},Losses_p_{name}", real=True)

    # auxiliar
    v_t = V_t_m*sym.exp(sym.I*theta_t)
    v_s = V_s_m*sym.exp(sym.I*theta_s)
    i_s = i_s_r + sym.I*i_s_i
    Z_s = R_s + sym.I*X_s
    
    KVL = v_t - i_s*Z_s - v_s  # Kirchhoff's voltage law for the AC side (bus_ac)
                
    # dynamic equations            


    # algebraic equations   
    g_i_s_r = sym.re(KVL)
    g_i_s_i = sym.im(KVL)

    # dae 
    grid.dae['f'] += []
    grid.dae['x'] += []
    grid.dae['g'] += [g_i_s_r,g_i_s_i]
    grid.dae['y_ini'] += [i_s_r,i_s_i]  
    grid.dae['y_run'] += [i_s_r,i_s_i]  
    
    grid.dae['u_ini_dict'].update({f'{V_t_m}':1.0})
    grid.dae['u_ini_dict'].update({f'{theta_t}':0.0})
    grid.dae['u_run_dict'].update({f'{V_t_m}':1.0})
    grid.dae['u_run_dict'].update({f'{theta_t}':0.0})         

    # parameters            
    grid.dae['params_dict'].update({f"{S_n}":data_dict['S_n']}) 
    grid.dae['params_dict'].update({f"{R_s}":data_dict['R_s']}) 
    grid.dae['params_dict'].update({f"{X_s}":data_dict['X_s']}) 

    Losses_p = 0.002 
    Losses_s = 0.01
    if 'Losses_p' in data_dict: Losses_p = data_dict['Losses_p']
    if 'Losses_s' in data_dict: Losses_s = data_dict['Losses_s']
    grid.dae['params_dict'].update({f"{Losses_p}":Losses_p}) 
    grid.dae['params_dict'].update({f"{Losses_s}":Losses_s}) 

    s_s = v_s*sym.conjugate(i_s)
    p_s = sym.re(s_s)
    q_s = sym.im(s_s)

    p_dc = -p_s + sym.Abs(s_s)**2*Losses_s + Losses_p

               
    # outputs
    grid.dae['h_dict'].update({f"i_s_m_{name}":sym.Abs(i_s_r + sym.I*i_s_i)})
    grid.dae['h_dict'].update({f"V_t_m_{name}":V_t_m})
    grid.dae['h_dict'].update({f"p_dc_{name}":p_dc})
    

    p_W_j   =  -p_dc*S_n
    q_var_j =  0.0

    p_W_k   = -p_s*S_n
    q_var_k = -q_s*S_n

    return p_W_j,q_var_j,p_W_k,q_var_k


def test():
    
    from pydae.bmapu import bmapu_builder
    import pytest

    grid = bmapu_builder.bmapu('vsc_acgf_dcp.hjson')
    #grid.checker()
    grid.verbose = True 
    grid.build('temp')

    import temp

    model = temp.model()

    model.ini({'P_ac':-1000e6},'xy_0.json')

    model.report_y()
    model.report_z()

    # assert model.get_value('i_s_m_2') == pytest.approx(soc_ref, rel=0.001)



if __name__=='__main__':

    test()


# -*- coding: utf-8 -*-
"""
Created on Thu August 10 23:52:55 2022

@author: jmmauricio
"""

import numpy as np
import sympy as sym

def vsc_acq_dcgf(grid,name,bus_j,bus_ac,data_dict):
    """
    # auxiliar
    
    bus_j: bus DC
    bus_ac: bus AC

    """

    # inputs from grid
    V_s_m   = sym.Symbol(f"V_{bus_ac}", real=True)
    theta_s = sym.Symbol(f"theta_{bus_ac}", real=True)
    V_d_m   = sym.Symbol(f"V_{bus_j}", real=True)
    theta_d = sym.Symbol(f"theta_{bus_j}", real=True)  # should be zero

    # inputs
    V_t_m = sym.Symbol(f"V_t_m_{bus_j}", real=True)     
    theta_t = sym.Symbol(f"theta_t_{bus_j}", real=True) # should be zero
    q_s_ref = sym.Symbol(f"q_s_ref_{bus_ac}", real=True) # should be zero
 
    # algebraic states
    i_d_r = sym.symbols(f"i_d_r_{bus_j}", real=True)
    i_d_i = sym.symbols(f"i_d_i_{bus_j}", real=True)


    # parameters
    S_n = sym.Symbol(f"S_n_{name}", real=True)
    R_d = sym.Symbol(f"R_d_{name}", real=True)
    X_d = sym.Symbol(f"X_d_{name}", real=True)
    Losses_s,Losses_p = sym.symbols(f"Losses_s_{name},Losses_p_{name}", real=True)

    # auxiliar
    v_t = V_t_m  # DC-side internal complex voltage (should be only real)
    v_d = V_d_m*sym.exp(sym.I*theta_d) # DC-side POI complex voltage (should be only real)
    v_s = V_s_m*sym.exp(sym.I*theta_s)  # AC-side POI complex voltage
    i_d = i_d_r + sym.I*i_d_i        # DC-side current (should be only real)
    Z_d = R_d               # DC-side impedance (should be only real)
    
    KVL = v_t - i_d*Z_d - v_d  # Kirchhoff's voltage law for the DC side (bus_j)
                
    # dynamic equations            


    # algebraic equations   
    g_i_d_r =  sym.re(KVL)
    g_i_d_i = i_d_i - sym.im(v_d)/1000.0

    # dae 
    grid.dae['f'] += []
    grid.dae['x'] += []
    grid.dae['g'] += [g_i_d_r,g_i_d_i]
    grid.dae['y_ini'] += [i_d_r,i_d_i]  
    grid.dae['y_run'] += [i_d_r,i_d_i]  
    
    grid.dae['u_ini_dict'].update({f'{V_t_m}':1.0})
    grid.dae['u_ini_dict'].update({f'{q_s_ref}':0.0})

    grid.dae['u_run_dict'].update({f'{V_t_m}':1.0})
    grid.dae['u_run_dict'].update({f'{q_s_ref}':0.0})         
               
    # outputs
    grid.dae['h_dict'].update({f"i_d_{name}":i_d_r})
    grid.dae['h_dict'].update({f"V_t_m_{name}":V_t_m})
    
    # parameters            
    grid.dae['params_dict'].update({f"{S_n}":data_dict['S_n']}) 
    grid.dae['params_dict'].update({f"{R_d}":data_dict['R_d']}) 
    grid.dae['params_dict'].update({f"{X_d}":data_dict['X_d']}) 

    Losses_p = 0.002 
    Losses_s = 0.01
    if 'Losses_p' in data_dict: Losses_p = data_dict['Losses_p']
    if 'Losses_s' in data_dict: Losses_s = data_dict['Losses_s']
    grid.dae['params_dict'].update({f"{Losses_p}":Losses_p}) 
    grid.dae['params_dict'].update({f"{Losses_s}":Losses_s}) 
    

    s_d = v_d*i_d
    p_d = sym.re(s_d)
    q_d = sym.im(s_d)

    s_s_m = sym.sqrt(p_d**2 + q_s_ref**2)

    p_s = -p_d + s_s_m**2*Losses_s + Losses_p
    q_s = q_s_ref

    grid.dae['h_dict'].update({f"P_dc_{name}":p_d})


    p_W_j   = -p_d*S_n  # DC side
    q_var_j = -q_d*S_n  # DC side

    p_W_k   = -p_s*S_n   # AC side
    q_var_k = -q_s*S_n

    return p_W_j,q_var_j,p_W_k,q_var_k


def test():
    
    from pydae.bmapu import bmapu_builder
    import pytest

    grid = bmapu_builder.bmapu('vsc_acq_dcgf.hjson')
    #grid.checker()
    grid.verbose = True 
    grid.build('temp')

    import temp

    model = temp.model()

    model.ini({'P_dc':-1000e6},'xy_0.json')

    model.report_y()
    model.report_z()

    # assert model.get_value('i_s_m_2') == pytest.approx(soc_ref, rel=0.001)

def test_hvdc():
    
    from pydae.bmapu import bmapu_builder
    import pytest

    grid = bmapu_builder.bmapu('hvdc.hjson')
    #grid.checker()
    grid.verbose = True 
    grid.build('temp')

    import temp

    model = temp.model()

    model.ini({'P_WFM':1500e6},'xy_0.json')

    model.report_y()
    model.report_z()

    p_GRI = model.get_value('p_s_GRI')*model.get_value('S_n_GRI')
    P_dc = model.get_value('p_dc_WFD_WFH')*model.get_value('S_n_WFD_WFH')

    print(f'P_GRI = {p_GRI/1e6:5.3f} MW')
    print(f'P_dc = {P_dc/1e6:5.3f} MW')

if __name__=='__main__':

    #test()
    test_hvdc()


import numpy as np
import sympy as sym


def ac3ph4wvdcq(grid,vsc_data):
    '''
    Converter type v_dc,q_ac 3 phase 4 wire
    
    '''
    
    bus_ac_name = vsc_data['bus_ac']
    bus_dc_name = vsc_data['bus_dc']  
    to_bus_dc_name = vsc_data['to_bus_dc']  
    
    a_value  = vsc_data['a']   
    b_value  = vsc_data['b']   
    c_value  = vsc_data['c']   
    
   
    ### AC-side
    q_a,q_b,q_c = sym.symbols(f'q_vsc_a_{bus_name}_a,q_vsc_b_{bus_name}_b,q_vsc_c_{bus_name}_c')

    p_ac,q_ac,p_dc,p_loss = sym.symbols(f'p_vsc_{bus_ac_name},q_vsc_{bus_ac_name},p_vsc_{bus_dc_name},p_vsc_loss_{bus_ac_name}',real=True)
    p_a_d,p_b_d,p_c_d,p_n_d = sym.symbols(f'p_a_d_{bus_ac_name},p_b_d_{bus_ac_name},p_c_d_{bus_ac_name},p_n_d_{bus_ac_name}',real=True)
    C_a,C_b,C_c = sym.symbols(f'C_a_{bus_ac_name},C_b_{bus_ac_name},C_c_{bus_ac_name}',real=True)
   
    #### AC voltages:
    v_a_r,v_a_i = sym.symbols(f'v_{bus_ac_name}_a_r,v_{bus_ac_name}_a_i',real=True)
    v_b_r,v_b_i = sym.symbols(f'v_{bus_ac_name}_b_r,v_{bus_ac_name}_b_i',real=True)
    v_c_r,v_c_i = sym.symbols(f'v_{bus_ac_name}_c_r,v_{bus_ac_name}_c_i',real=True)
    v_n_r,v_n_i = sym.symbols(f'v_{bus_ac_name}_n_r,v_{bus_ac_name}_n_i',real=True)
    
    #### AC currents:
    i_a_r,i_a_i = sym.symbols(f'i_vsc_{bus_ac_name}_a_r,i_vsc_{bus_ac_name}_a_i',real=True)
    i_b_r,i_b_i = sym.symbols(f'i_vsc_{bus_ac_name}_b_r,i_vsc_{bus_ac_name}_b_i',real=True)
    i_c_r,i_c_i = sym.symbols(f'i_vsc_{bus_ac_name}_c_r,i_vsc_{bus_ac_name}_c_i',real=True)
    i_n_r,i_n_i = sym.symbols(f'i_vsc_{bus_ac_name}_n_r,i_vsc_{bus_ac_name}_n_i',real=True)
    
    # DC line current for computing DC power
    i_l_a_r,i_l_n_r = sym.symbols(f'i_l_{bus_dc_name}_{to_bus_dc_name}_a_r,i_l_{bus_dc_name}_{to_bus_dc_name}_n_r', real = True)   
    v_dc_a_r,v_dc_n_r  = sym.symbols(f'v_{bus_dc_name}_a_r,v_{bus_dc_name}_n_r', real = True) 
    v_dc_ref  = sym.Symbol(f'v_dc_{bus_dc_name}', real = True) 

    a,b,c,K_dc = sym.symbols(f'a_{bus_ac_name},b_{bus_ac_name},c_{bus_ac_name},K_dc_{bus_dc_name}',real=True)
    
    i_a_rms = sym.sqrt(i_a_r**2+i_a_i**2 + 0.01) 
    i_b_rms = sym.sqrt(i_b_r**2+i_b_i**2+ 0.01) 
    i_c_rms = sym.sqrt(i_c_r**2+i_c_i**2+ 0.01) 
    i_n_rms = sym.sqrt(i_n_r**2+i_n_i**2+ 0.01) 


    p_loss_a_ = a + b*i_a_rms + c*i_a_rms*i_a_rms
    p_loss_b_ = a + b*i_b_rms + c*i_b_rms*i_b_rms
    p_loss_c_ = a + b*i_c_rms + c*i_c_rms*i_c_rms
    p_loss_n_ = a + b*i_n_rms + c*i_n_rms*i_n_rms
      
    

    v_a = v_a_r + 1j*v_a_i
    v_b = v_b_r + 1j*v_b_i
    v_c = v_c_r + 1j*v_c_i
    v_n = v_n_r + 1j*v_n_i

    i_a = i_a_r + 1j*i_a_i
    i_b = i_b_r + 1j*i_b_i
    i_c = i_c_r + 1j*i_c_i
    i_n = i_n_r + 1j*i_n_i


    s_a = (v_a - v_n) * sym.conjugate(i_a)
    s_b = (v_b - v_n) * sym.conjugate(i_b)
    s_c = (v_c - v_n) * sym.conjugate(i_c)
    s_n = (v_n)* sym.conjugate(i_n)
    
    
    eq_p_a_d =  C_a*p_dc - p_a_d 
    eq_p_b_d =  C_b*p_dc - p_b_d
    eq_p_c_d =  C_c*p_dc - p_c_d
    eq_p_n_d =  sym.re(s_n) - p_n_d
    
    #p_loss_a = sym.Piecewise((-p_loss_a_, p_a_d < 0), (p_loss_a_, p_a_d > 0),(p_loss_a_, True))
    #p_loss_b = sym.Piecewise((-p_loss_b_, p_b_d < 0), (p_loss_b_, p_b_d > 0),(p_loss_b_, True))
    #p_loss_c = sym.Piecewise((-p_loss_c_, p_c_d < 0), (p_loss_c_, p_c_d > 0),(p_loss_c_, True))
    #p_loss_n = sym.Piecewise((-p_loss_n_, p_n_d < 0), (p_loss_n_, p_n_d > 0),(p_loss_n_, True))

    eq_i_a_r =  sym.re(s_a) - p_a_d + p_loss_a_ + p_loss_n_
    eq_i_b_r =  sym.re(s_b) - p_b_d + p_loss_b_
    eq_i_c_r =  sym.re(s_c) - p_c_d + p_loss_c_
    eq_i_a_i =  sym.im(s_a) - q_c
    eq_i_b_i =  sym.im(s_b) - q_b
    eq_i_c_i =  sym.im(s_c) - q_c
    
    eq_i_n_r = -i_n_r + i_a_r + i_b_r + i_c_r
    eq_i_n_i = -i_n_i + i_a_i + i_b_i + i_c_i
   
    eq_p_dc = p_dc + i_l_a_r * v_dc_a_r + i_l_n_r * v_dc_n_r    
    eq_v_dc_a_r = -v_dc_a_r + v_dc_ref + v_dc_n_r - K_dc * p_dc


    grid.dae['g'] += [eq_p_a_d,
                      eq_p_b_d,
                      eq_p_c_d,
                      eq_p_n_d,
                      eq_i_a_r,eq_i_a_i,
                      eq_i_b_r,eq_i_b_i,
                      eq_i_c_r,eq_i_c_i,
                      eq_i_n_r,eq_i_n_i,
                      eq_p_dc,
                      eq_v_dc_a_r
                      ]
    
    grid.dae['y'] += [p_a_d,
                      p_b_d,
                      p_c_d,
                      p_n_d,
                      i_a_r,   i_a_i,
                      i_b_r,   i_b_i,
                      i_c_r,   i_c_i,
                      i_n_r,   i_n_i,
                      p_dc,
                      v_dc_a_r]
    
    i_abc_list  = [i_a_r,i_a_i,i_b_r,i_b_i,i_c_r,i_c_i]
    for itg in [1,2,3]:
        bus_idx = grid.nodes.index(f'{bus_ac_name}.{itg}')
        g_idx = bus_idx - grid.N_nodes_v
        grid.dae['g'][2*g_idx+0] += i_abc_list[2*(itg-1)  ]
        grid.dae['g'][2*g_idx+1] += i_abc_list[2*(itg-1)+1]

    bus_idx = grid.nodes.index(f'{bus_dc_name}.{4}')
    g_idx = bus_idx - grid.N_nodes_v
    grid.dae['g'][2*g_idx+0] += -i_l_a_r
    grid.dae['g'][2*g_idx+1] +=        0   

    
      
    grid.dae['u'].update({f'v_dc_{bus_dc_name}':800.0,f'{str(q_a)}':0.0,f'{str(q_b)}':0.0,f'{str(q_c)}':0.0}) 
    grid.dae['u'].pop(str(v_dc_a_r))
    grid.dae['params'].update({f'a_{bus_ac_name}':a_value,f'b_{bus_ac_name}':b_value,f'c_{bus_ac_name}':c_value})
    grid.dae['params'].update({f'C_a_{bus_ac_name}':1/3,f'C_b_{bus_ac_name}':1/3,f'C_c_{bus_ac_name}':1/3})
    grid.dae['params'].update({f'K_dc_{bus_dc_name}':1e-6})

    grid.dae['xy_0_dict'].update({f'v_{bus_dc_name}_a_r':800.0,f'v_{bus_dc_name}_n_r':1.0})
    
    grid.dae['h_dict'].update({f'p_vsc_{bus_ac_name}':sym.re(s_a)+sym.re(s_b)+sym.re(s_c)+sym.re(s_n)})
    grid.dae['h_dict'].update({f'p_vsc_loss_{bus_ac_name}':(p_loss_a_+p_loss_b_+p_loss_c_+p_loss_n_)})
    



import numpy as np
import scipy.sparse as sspa
import cffi
import solver_ini,solver_run

dae_file_mode = 'local'

ffi = cffi.FFI()

sparse = False


if sparse:
    sp_jac_ini_xy_eval = jacs.lib.sp_jac_ini_xy_eval
    sp_jac_ini_up_eval = jacs.lib.sp_jac_ini_up_eval
    sp_jac_ini_num_eval = jacs.lib.sp_jac_ini_num_eval


if sparse:
    sp_jac_run_xy_eval = jacs.lib.sp_jac_run_xy_eval
    sp_jac_run_up_eval = jacs.lib.sp_jac_run_up_eval
    sp_jac_run_num_eval = jacs.lib.sp_jac_run_num_eval


if sparse:
    sp_jac_trap_xy_eval= jacs.lib.sp_jac_trap_xy_eval            
    sp_jac_trap_up_eval= jacs.lib.sp_jac_trap_up_eval        
    sp_jac_trap_num_eval= jacs.lib.sp_jac_trap_num_eval





import json

sin = np.sin
cos = np.cos
atan2 = np.arctan2
sqrt = np.sqrt 
sign = np.sign 
exp = np.exp


class model: 

    def __init__(self): 
        
        self.matrices_folder = 'build'
        self.sparse = True
        self.dae_file_mode = 'local'
        self.t_end = 10.000000 
        self.Dt = 0.0010000 
        self.decimation = 10.000000 
        self.itol = 1e-6 
        self.Dt_max = 0.001000 
        self.Dt_min = 0.001000 
        self.solvern = 5 
        self.imax = 100 
        self.N_x = 4
        self.N_y = 2 
        self.N_z = 4 
        self.N_store = 100000 
        self.params_list = ['L', 'G', 'M', 'K_d'] 
        self.params_values_list  = [5.21, 9.81, 10.0, 0.001] 
        self.inputs_ini_list = ['theta', 'u_dummy'] 
        self.inputs_ini_values_list  = [0.08726646259971647, 0.0] 
        self.inputs_run_list = ['f_x', 'u_dummy'] 
        self.inputs_run_values_list = [0, 0.0] 
        self.outputs_list = ['E_p', 'E_k', 'f_x', 'lam'] 
        self.x_list = ['p_x', 'p_y', 'v_x', 'v_y'] 
        self.y_run_list = ['lam', 'theta'] 
        self.xy_list = self.x_list + self.y_run_list 
        self.y_ini_list = ['lam', 'f_x'] 
        self.xy_ini_list = self.x_list + self.y_ini_list 
        self.t = 0.0
        self.it = 0
        self.it_store = 0
        self.xy_prev = np.zeros((self.N_x+self.N_y,1))
        self.initialization_tol = 1e-6
        self.N_u = len(self.inputs_run_list) 
        self.sopt_root_method='hybr'
        self.sopt_root_jac=True
        self.u_ini_list = self.inputs_ini_list
        self.u_ini_values_list = self.inputs_ini_values_list
        self.u_run_list = self.inputs_run_list
        self.u_run_values_list = self.inputs_run_values_list
        self.N_u = len(self.u_run_list)
        self.u_ini = np.array(self.inputs_ini_values_list, dtype=np.float64)
        self.p = np.array(self.params_values_list, dtype=np.float64)
        self.xy_0 = np.zeros((self.N_x+self.N_y,),dtype=np.float64)
        self.xy = np.zeros((self.N_x+self.N_y,),dtype=np.float64)
        self.z = np.zeros((self.N_z,),dtype=np.float64)
        
        # numerical elements of jacobians computing:
        x = self.xy[:self.N_x]
        y = self.xy[self.N_x:]
        
        self.yini2urun = list(set(self.u_run_list).intersection(set(self.y_ini_list)))
        self.uini2yrun = list(set(self.y_run_list).intersection(set(self.u_ini_list)))
        self.Time = np.zeros(self.N_store)
        self.X = np.zeros((self.N_store,self.N_x))
        self.Y = np.zeros((self.N_store,self.N_y))
        self.Z = np.zeros((self.N_store,self.N_z))
        self.iters = np.zeros(self.N_store) 
        self.u_run = np.array(self.u_run_values_list,dtype=np.float64)
 
        ## jac_ini
        if self.sparse:
            self.sp_jac_ini_indices, self.sp_jac_ini_indptr, self.sp_jac_ini_nia, self.sp_jac_ini_nja = sp_jac_ini_vectors()
            self.sp_jac_ini_indices = np.array(self.sp_jac_ini_indices,dtype=np.int32)    
            self.sp_jac_ini_indptr = np.array(self.sp_jac_ini_indptr,dtype=np.int32)    
            self.sp_jac_ini_data = np.array(self.sp_jac_ini_indices,dtype=np.float64)            

        ## jac_run
        if self.sparse:
            self.sp_jac_run_indices, self.sp_jac_run_indptr, self.sp_jac_run_nia, self.sp_jac_run_nja = sp_jac_run_vectors()
            self.sp_jac_run_indices = np.array(self.sp_jac_run_indices,dtype=np.int32)    
            self.sp_jac_run_indptr = np.array(self.sp_jac_run_indptr,dtype=np.int32)    
            self.sp_jac_run_data = np.array(self.sp_jac_run_indices,dtype=np.float64)

        ## jac_trap
        if self.sparse:
            self.sp_jac_trap_indices, self.sp_jac_trap_indptr, self.sp_jac_trap_nia, self.sp_jac_trap_nja = sp_jac_trap_vectors()
            self.sp_jac_trap_indices = np.array(self.sp_jac_trap_indices,dtype=np.int32)    
            self.sp_jac_trap_indptr = np.array(self.sp_jac_trap_indptr,dtype=np.int32)    
            self.sp_jac_trap_data = np.array(self.sp_jac_trap_indices,dtype=np.float64)
        
        self.max_it,self.itol,self.store = 50,1e-8,1 
        self.lmax_it,self.ltol,self.ldamp= 50,1e-8,1.0
        self.mode = 0 

        self.lmax_it_ini,self.ltol_ini,self.ldamp_ini=50,1e-8,1.0

        #self.sp_Fu_run = sspa.load_npz(f'./{self.matrices_folder}/temp_Fu_run_num.npz')
        #self.sp_Gu_run = sspa.load_npz(f'./{self.matrices_folder}/temp_Gu_run_num.npz')
        #self.sp_Hx_run = sspa.load_npz(f'./{self.matrices_folder}/temp_Hx_run_num.npz')
        #self.sp_Hy_run = sspa.load_npz(f'./{self.matrices_folder}/temp_Hy_run_num.npz')
        #self.sp_Hu_run = sspa.load_npz(f'./{self.matrices_folder}/temp_Hu_run_num.npz')        
        
        self.ss_solver = 2
        self.lsolver = 2

        # ini initialization
        self.inidblparams = np.zeros(10,dtype=np.float64)
        self.iniintparams = np.zeros(10,dtype=np.int32)

        # run initialization
        self.rundblparams = np.zeros(10,dtype=np.float64)
        self.runintparams = np.zeros(10,dtype=np.int32)

        self.xy = self.xy_0
        self.pt_ini = np.zeros(64,dtype=np.int32)
        xy = self.xy
        x = xy[:self.N_x]
        y_ini = xy[self.N_x:]
        Dxy = np.zeros((self.N_x+self.N_y),dtype=np.float64)
        
        self.f = np.zeros((self.N_x),dtype=np.float64)
        self.g = np.zeros((self.N_y),dtype=np.float64)
        self.fg = np.zeros((self.N_x+self.N_y),dtype=np.float64)


        self.p_pt_ini =solver_ini.ffi.cast('int *', self.pt_ini.ctypes.data)
        self.p_sp_jac_ini = solver_ini.ffi.cast('double *', self.sp_jac_ini_data.ctypes.data)
        self.p_indptr = solver_ini.ffi.cast('int *', self.sp_jac_ini_indptr.ctypes.data)
        self.p_indices = solver_ini.ffi.cast('int *', self.sp_jac_ini_indices.ctypes.data)
        self.p_x = solver_ini.ffi.cast('double *', x.ctypes.data)
        self.p_y_ini = solver_ini.ffi.cast('double *', y_ini.ctypes.data)
        self.p_xy = solver_ini.ffi.cast('double *', self.xy.ctypes.data)
        self.p_Dxy = solver_ini.ffi.cast('double *', Dxy.ctypes.data)
        self.p_u_ini = solver_ini.ffi.cast('double *', self.u_ini.ctypes.data)
        self.p_p = solver_ini.ffi.cast('double *', self.p.ctypes.data)
        self.p_z = solver_ini.ffi.cast('double *', self.z.ctypes.data)
        self.p_inidblparams = solver_ini.ffi.cast('double *', self.inidblparams.ctypes.data)
        self.p_iniintparams = solver_ini.ffi.cast('int *', self.iniintparams.ctypes.data)
        self.p_f = solver_ini.ffi.cast('double *', self.f.ctypes.data)
        self.p_g = solver_ini.ffi.cast('double *', self.g.ctypes.data)
        self.p_fg = solver_ini.ffi.cast('double *', self.fg.ctypes.data)

        self.pt_run = np.zeros(64,dtype=np.int32)

        self.p_pt_run =solver_ini.ffi.cast('int *', self.pt_run.ctypes.data)

        self.X = np.zeros((100_000*self.N_x))
        self.Y = np.zeros((100_000*self.N_y))
        self.Z = np.zeros((100_000*self.N_z))
        self.Time = np.zeros((100_000))
        self.its = np.array([0],dtype=np.int32)

    # def update(self):

    #     self.Time = np.zeros(self.N_store)
    #     self.X = np.zeros((self.N_store,self.N_x))
    #     self.Y = np.zeros((self.N_store,self.N_y))
    #     self.Z = np.zeros((self.N_store,self.N_z))
    #     self.iters = np.zeros(self.N_store)
    
    def jac_run_eval(self):
        de_jac_run_eval(self.jac_run,self.x,self.y_run,self.u_run,self.p,self.Dt)
      
    def post_run_checker(self):

        self.N_iters = self.runintparams[2]  # number of iterations done
        self.N_mkl_factorizations = self.runintparams[3]
        self.N_mkl_resets = self.runintparams[4]
        self.pt_run_sum = np.sum(self.pt_run) # this variable should be 0

    def run(self,t_end,up_dict):

        for item in up_dict:
            self.set_value(item,up_dict[item])
            
        t = self.t
        it = self.it
        it_store = self.it_store
        xy = self.xy
        u = self.u_run
        z = self.z

        pt = np.zeros(64,dtype=np.int32)
        xy = self.xy
        x = xy[:self.N_x]
        y_run = xy[self.N_x:]

        p_sp_jac_trap = solver_run.ffi.cast('double *', self.sp_jac_trap_data.ctypes.data)
        p_indptr = solver_run.ffi.cast('int *', self.sp_jac_trap_indptr.ctypes.data)
        p_indices = solver_run.ffi.cast('int *', self.sp_jac_trap_indices.ctypes.data)
        p_x = solver_run.ffi.cast('double *', x.ctypes.data)
        p_y_run = solver_run.ffi.cast('double *', y_run.ctypes.data)
        p_xy = solver_run.ffi.cast('double *', self.xy.ctypes.data)
        p_u_run = solver_run.ffi.cast('double *', self.u_run.ctypes.data)
        p_z = solver_run.ffi.cast('double *', self.z.ctypes.data)
        p_dblparams = solver_run.ffi.cast('double *', self.rundblparams.ctypes.data)
        p_intparams = solver_run.ffi.cast('int *', self.runintparams.ctypes.data)
        p_x = solver_run.ffi.cast('double *', x.ctypes.data)
        p_X = solver_run.ffi.cast('double *', self.X.ctypes.data)
        p_Y = solver_run.ffi.cast('double *', self.Y.ctypes.data)
        p_Z = solver_run.ffi.cast('double *', self.Z.ctypes.data)
        p_Time = solver_run.ffi.cast('double *', self.Time.ctypes.data)
        p_its = solver_run.ffi.cast('int *', self.its.ctypes.data)
        p_p = solver_run.ffi.cast('double *', self.p.ctypes.data)
        N_x = self.N_x
        N_y = self.N_y
        max_it = self.max_it
        itol = self.itol
        max_it = 5
        itol = 1e-8

        solver_run.lib.run(self.p_pt_run, t, t_end,p_sp_jac_trap, p_indptr,p_indices,p_x,p_y_run,p_xy,p_u_run,p_p,N_x,N_y, max_it, itol, p_its, self.Dt, p_z,p_dblparams, p_intparams, p_Time, p_X, p_Y, p_Z, self.N_z, self.N_store)


        self.t = self.rundblparams[0]
        self.it = it
        self.it_store = self.its[0]
        self.xy = xy
        self.z = z

    def step(self,t_end,up_dict):

        for item in up_dict:
            self.set_value(item,up_dict[item])
            
        t = self.t
        it = self.it
        it_store = self.it_store
        xy = self.xy
        u = self.u_run
        z = self.z

        pt = np.zeros(64,dtype=np.int32)
        xy = self.xy
        x = xy[:self.N_x]
        y_run = xy[self.N_x:]

        p_sp_jac_trap = solver_run.ffi.cast('double *', self.sp_jac_trap_data.ctypes.data)
        p_indptr = solver_run.ffi.cast('int *', self.sp_jac_trap_indptr.ctypes.data)
        p_indices = solver_run.ffi.cast('int *', self.sp_jac_trap_indices.ctypes.data)
        p_x = solver_run.ffi.cast('double *', x.ctypes.data)
        p_y_run = solver_run.ffi.cast('double *', y_run.ctypes.data)
        p_xy = solver_run.ffi.cast('double *', self.xy.ctypes.data)
        p_u_run = solver_run.ffi.cast('double *', self.u_run.ctypes.data)
        p_z = solver_run.ffi.cast('double *', self.z.ctypes.data)
        p_dblparams = solver_run.ffi.cast('double *', self.rundblparams.ctypes.data)
        p_intparams = solver_run.ffi.cast('int *', self.runintparams.ctypes.data)
        p_x = solver_run.ffi.cast('double *', x.ctypes.data)
        p_X = solver_run.ffi.cast('double *', self.X.ctypes.data)
        p_Y = solver_run.ffi.cast('double *', self.Y.ctypes.data)
        p_Z = solver_run.ffi.cast('double *', self.Z.ctypes.data)
        p_Time = solver_run.ffi.cast('double *', self.Time.ctypes.data)
        p_its = solver_run.ffi.cast('int *', self.its.ctypes.data)
        p_p = solver_run.ffi.cast('double *', self.p.ctypes.data)
        N_x = self.N_x
        N_y = self.N_y
        max_it = self.max_it
        itol = self.itol
        itol = 1e-8

        solver_run.lib.step3(self.p_pt_run, t, t_end,p_sp_jac_trap, p_indptr,p_indices,p_x,p_y_run,p_xy,p_u_run,p_p,N_x,N_y, max_it, itol, p_its, self.Dt, p_z,p_dblparams, p_intparams, p_Time, p_X, p_Y, p_Z, self.N_z, self.N_store)

        self.t = self.rundblparams[0]
        self.it = it
        self.it_store = self.its[0]
        self.xy = xy
        self.z = z
        self.it_max = self.runintparams[5]
        
    def post(self):
        
        self.Time = self.Time[:self.its[0]].reshape(self.its[0],1)
        self.X = self.X[:self.its[0]*self.N_x].reshape(self.its[0],self.N_x)
        self.Y = self.Y[:self.its[0]*self.N_y].reshape(self.its[0],self.N_y)
        self.Z = self.Z[:self.its[0]*self.N_z].reshape(self.its[0],self.N_z)

        
    def ini2run(self):
        
        ## y_ini to y_run
        self.y_ini = self.xy_ini[self.N_x:]
        self.y_run = np.copy(self.y_ini)
        
        ## y_ini to u_run
        for item in self.yini2urun:
            self.u_run[self.u_run_list.index(item)] = self.y_ini[self.y_ini_list.index(item)]
                
        ## u_ini to y_run
        for item in self.uini2yrun:
            self.y_run[self.y_run_list.index(item)] = self.u_ini[self.u_ini_list.index(item)]
            
        
        self.x = self.xy_ini[:self.N_x]
        self.xy[:self.N_x] = self.x
        self.xy[self.N_x:] = self.y_run        

    def get_value(self,name):
        
        if name in self.inputs_run_list:
            value = self.u_run[self.inputs_run_list.index(name)]
            return value
            
        if name in self.x_list:
            idx = self.x_list.index(name)
            value = self.xy[idx]
            return value
            
        if name in self.y_run_list:
            idy = self.y_run_list.index(name)
            value = self.xy[self.N_x+idy]
            return value
        
        if name in self.params_list:
            idp = self.params_list.index(name)
            value = self.p[idp]
            return value
            
        if name in self.outputs_list:
            idz = self.outputs_list.index(name)
            value = self.z[idz]
            return value

    def get_values(self,name):
        if name in self.x_list:
            values = self.X[:,self.x_list.index(name)]
        if name in self.y_run_list:
            values = self.Y[:,self.y_run_list.index(name)]
        if name in self.outputs_list:
            values = self.Z[:,self.outputs_list.index(name)]
                        
        return values

    def get_mvalue(self,names):
        '''

        Parameters
        ----------
        names : list
            list of variables names to return each value.

        Returns
        -------
        mvalue : TYPE
            list of value of each variable.

        '''
        mvalue = []
        for name in names:
            mvalue += [self.get_value(name)]
                        
        return mvalue
    
    def set_value(self,name_,value):
        if name_ in self.inputs_ini_list or name_ in self.inputs_run_list:
            if name_ in self.inputs_ini_list:
                self.u_ini[self.inputs_ini_list.index(name_)] = value
            if name_ in self.inputs_run_list:
                self.u_run[self.inputs_run_list.index(name_)] = value
            return
        elif name_ in self.params_list:
            self.p[self.params_list.index(name_)] = value
            return
        else:
            print(f'Input or parameter {name_} not found.')
 
    def report_x(self,value_format='5.2f'):
        for item in self.x_list:
            print(f'{item:5s} = {self.get_value(item):{value_format}}')

    def report_y(self,value_format='5.2f'):
        for item in self.y_run_list:
            print(f'{item:5s} = {self.get_value(item):{value_format}}')
            
    def report_u(self,value_format='5.2f'):
        for item in self.inputs_run_list:
            print(f'{item:5s} ={self.get_value(item):{value_format}}')

    def report_z(self,value_format='5.2f'):
        for item in self.outputs_list:
            print(f'{item:5s} = {self.get_value(item):{value_format}}')

    def report_params(self,value_format='5.2f'):
        for item in self.params_list:
            print(f'{item:5s} ={self.get_value(item):{value_format}}')


    def post_ini_checker(self):

        self.N_iters = self.iniintparams[2]  # number of iterations done
        self.N_mkl_factorizations = self.iniintparams[3]
        self.N_mkl_resets = self.iniintparams[4]
        self.pt_ini_sum = np.sum(self.pt_ini) # this variable should be 0


    def ini(self,up_dict,xy_0={}):
        '''
        Find the steady state of the initialization problem:
            
               0 = f(x,y,u,p) 
               0 = g(x,y,u,p) 

        Parameters
        ----------
        up_dict : dict
            dictionary with all the parameters p and inputs u new values.
        xy_0: if scalar, all the x and y values initial guess are set to the scalar.
              if dict, the initial guesses are applied for the x and y that are in the dictionary
              if string, the initial guess considers a json file with the x and y names and their initial values

        Returns
        -------
        mvalue : TYPE
            list of value of each variable.

        '''
        
        self.it = 0
        self.it_store = 0
        self.t = 0.0
    
        for item in up_dict:
            self.set_value(item,up_dict[item])
            
        if type(xy_0) == dict:
            xy_0_dict = xy_0
            self.dict2xy0(xy_0_dict)
            
        if type(xy_0) == str:
            if xy_0 == 'eval':
                N_x = self.N_x
                self.xy_0_new = np.copy(self.xy_0)*0
                xy0_eval(self.xy_0_new[:N_x],self.xy_0_new[N_x:],self.u_ini,self.p)
                self.xy_0_evaluated = np.copy(self.xy_0_new)
                self.xy_0 = np.copy(self.xy_0_new)
            else:
                self.load_xy_0(file_name = xy_0)
                
        if type(xy_0) == float or type(xy_0) == int:
            self.xy_0 = np.ones(self.N_x+self.N_y,dtype=np.float64)*xy_0

        self.f = np.zeros((self.N_x),dtype=np.float64)
        self.g = np.zeros((self.N_y),dtype=np.float64)
        self.fg = np.zeros((self.N_x+self.N_y),dtype=np.float64)

        xy = self.xy_0
        x = xy[:self.N_x]
        y_ini = xy[self.N_x:]
        Dxy = np.zeros((self.N_x+self.N_y),dtype=np.float64)
        self.p_pt_ini =solver_ini.ffi.cast('int *', self.pt_ini.ctypes.data)
        self.p_sp_jac_ini = solver_ini.ffi.cast('double *', self.sp_jac_ini_data.ctypes.data)
        self.p_indptr = solver_ini.ffi.cast('int *', self.sp_jac_ini_indptr.ctypes.data)
        self.p_indices = solver_ini.ffi.cast('int *', self.sp_jac_ini_indices.ctypes.data)
        self.p_x = solver_ini.ffi.cast('double *', x.ctypes.data)
        self.p_y_ini = solver_ini.ffi.cast('double *', y_ini.ctypes.data)
        self.p_xy = solver_ini.ffi.cast('double *', self.xy.ctypes.data)
        self.p_Dxy = solver_ini.ffi.cast('double *', Dxy.ctypes.data)
        self.p_u_ini = solver_ini.ffi.cast('double *', self.u_ini.ctypes.data)
        self.p_p = solver_ini.ffi.cast('double *', self.p.ctypes.data)
        self.p_z = solver_ini.ffi.cast('double *', self.z.ctypes.data)
        self.p_inidblparams = solver_ini.ffi.cast('double *', self.inidblparams.ctypes.data)
        self.p_iniintparams = solver_ini.ffi.cast('int *', self.iniintparams.ctypes.data)
        self.p_f = solver_ini.ffi.cast('double *', self.f.ctypes.data)
        self.p_g = solver_ini.ffi.cast('double *', self.g.ctypes.data)
        self.p_fg = solver_ini.ffi.cast('double *', self.fg.ctypes.data)


        solver_ini.lib.ini(self.p_pt_ini,
                            self.p_sp_jac_ini,
                            self.p_indptr,
                            self.p_indices,
                            self.p_x,
                            self.p_y_ini,
                            self.p_xy,
                            self.p_Dxy,
                            self.p_u_ini,
                            self.p_p,
                            self.N_x,
                            self.N_y,
                            self.max_it,
                            self.itol,
                            self.p_z,
                            self.p_inidblparams,
                            self.p_iniintparams,
                            self.p_f,
                            self.p_g,
                            self.p_fg)
        
        self.post_ini_checker()
        
        if self.iniintparams[2] < self.max_it-1:
            
            self.xy_ini = self.xy
            self.N_iters = self.iniintparams[2]

            self.ini2run()
            
            self.ini_convergence = True
            
        if self.iniintparams[2] >= self.max_it-1:
            print(f'Maximum number of iterations (max_it = {self.max_it}) reached without convergence.')
            self.ini_convergence = False
            
        return self.ini_convergence
            
        
    def dict2xy0(self,xy_0_dict):
    
        for item in xy_0_dict:
            if item in self.x_list:
                self.xy_0[self.x_list.index(item)] = xy_0_dict[item]
            if item in self.y_ini_list:
                self.xy_0[self.y_ini_list.index(item) + self.N_x] = xy_0_dict[item]
        
    
    def save_xy_0(self,file_name = 'xy_0.json'):
        xy_0_dict = {}
        for item in self.x_list:
            xy_0_dict.update({item:self.get_value(item)})
        for item in self.y_ini_list:
            xy_0_dict.update({item:self.get_value(item)})
    
        xy_0_str = json.dumps(xy_0_dict, indent=4)
        with open(file_name,'w') as fobj:
            fobj.write(xy_0_str)
    
    def load_xy_0(self,file_name = 'xy_0.json'):
        with open(file_name) as fobj:
            xy_0_str = fobj.read()
        xy_0_dict = json.loads(xy_0_str)
    
        for item in xy_0_dict:
            if item in self.x_list:
                self.xy_0[self.x_list.index(item)] = xy_0_dict[item]
            if item in self.y_ini_list:
                self.xy_0[self.y_ini_list.index(item)+self.N_x] = xy_0_dict[item]            

    def load_params(self,data_input):
    
        if type(data_input) == str:
            json_file = data_input
            self.json_file = json_file
            self.json_data = open(json_file).read().replace("'",'"')
            data = json.loads(self.json_data)
        elif type(data_input) == dict:
            data = data_input
    
        self.data = data
        for item in self.data:
            self.set_value(item, self.data[item])

    def save_params(self,file_name = 'parameters.json'):
        params_dict = {}
        for item in self.params_list:
            params_dict.update({item:self.get_value(item)})

        params_dict_str = json.dumps(params_dict, indent=4)
        with open(file_name,'w') as fobj:
            fobj.write(params_dict_str)

    def save_inputs_ini(self,file_name = 'inputs_ini.json'):
        inputs_ini_dict = {}
        for item in self.inputs_ini_list:
            inputs_ini_dict.update({item:self.get_value(item)})

        inputs_ini_dict_str = json.dumps(inputs_ini_dict, indent=4)
        with open(file_name,'w') as fobj:
            fobj.write(inputs_ini_dict_str)


    def eval_jac_u2z(self):

        '''

        0 =   J_run * xy + FG_u * u
        z = Hxy_run * xy + H_u * u

        xy = -1/J_run * FG_u * u
        z = -Hxy_run/J_run * FG_u * u + H_u * u
        z = (-Hxy_run/J_run * FG_u + H_u ) * u 
        '''
        
        sp_Fu_run_eval(self.sp_Fu_run.data,self.x,self.y_run,self.u_run,self.p,self.Dt)
        sp_Gu_run_eval(self.sp_Gu_run.data,self.x,self.y_run,self.u_run,self.p,self.Dt)
        sp_H_jacs_run_eval(self.sp_Hx_run.data,
                        self.sp_Hy_run.data,
                        self.sp_Hu_run.data,
                        self.x,self.y_run,self.u_run,self.p,self.Dt)
        sp_jac_run = self.sp_jac_run
        sp_jac_run_eval(sp_jac_run.data,
                        self.x,self.y_run,
                        self.u_run,self.p,
                        self.Dt)



        Hxy_run = sspa.bmat([[self.sp_Hx_run,self.sp_Hy_run]])
        FGu_run = sspa.bmat([[self.sp_Fu_run],[self.sp_Gu_run]])
        

        #((sspa.linalg.spsolve(s.sp_jac_ini,-Hxy_run)) @ FGu_run + sp_Hu_run )@s.u_ini

        self.jac_u2z = Hxy_run @ sspa.linalg.spsolve(self.sp_jac_run,-FGu_run) + self.sp_Hu_run  
        
        
    # def step(self,t_end,up_dict):
    #     for item in up_dict:
    #         self.set_value(item,up_dict[item])

    #     t = self.t
    #     p = self.p
    #     it = self.it
    #     it_store = self.it_store
    #     xy = self.xy
    #     u = self.u_run
    #     z = self.z

    #     pt = np.zeros(64,dtype=np.int32)
    #     xy = self.xy
    #     x = xy[:self.N_x]
    #     y_run = xy[self.N_x:]

    #     p_pt =solver_run.ffi.cast('int *', pt.ctypes.data)
    #     p_sp_jac_trap = solver_run.ffi.cast('double *', self.sp_jac_trap_data.ctypes.data)
    #     p_indptr = solver_run.ffi.cast('int *', self.sp_jac_trap_indptr.ctypes.data)
    #     p_indices = solver_run.ffi.cast('int *', self.sp_jac_trap_indices.ctypes.data)
    #     p_x = solver_run.ffi.cast('double *', x.ctypes.data)
    #     p_y_run = solver_run.ffi.cast('double *', y_run.ctypes.data)
    #     p_xy = solver_run.ffi.cast('double *', self.xy.ctypes.data)
    #     p_u_run = solver_run.ffi.cast('double *', self.u_run.ctypes.data)
    #     p_z = solver_run.ffi.cast('double *', self.z.ctypes.data)
    #     p_dblparams = solver_run.ffi.cast('double *', self.rundblparams.ctypes.data)
    #     p_intparams = solver_run.ffi.cast('int *', self.runintparams.ctypes.data)
    #     p_x = solver_run.ffi.cast('double *', x.ctypes.data)
    #     p_its = solver_run.ffi.cast('int *', self.its.ctypes.data)
    #     p_p = solver_run.ffi.cast('double *', self.p.ctypes.data)
    #     N_x = self.N_x
    #     N_y = self.N_y
    #     max_it = self.max_it
    #     itol = self.itol
    #     max_it = 5
    #     itol = 1e-8

    #     N_x = self.N_x
    #     N_y = self.N_y
    #     max_it = self.max_it
    #     itol = self.itol
    #     max_it = 5
    #     itol = 1e-8
    #     its = 0

    # #   solver_run.lib.run( self.p_pt_run, t, t_end,p_sp_jac_trap, p_indptr,p_indices,p_x,p_y_run,p_xy,p_u_run,p_p,N_x,N_y, max_it, itol, p_its, self.Dt, p_z,p_dblparams, p_intparams, p_Time, p_X, p_Y, p_Z, self.N_z, self.N_store)
    #     solver_run.lib.step(self.p_pt_run, t, t_end,p_sp_jac_trap, p_indptr,p_indices,p_x,p_y_run,p_xy,p_u_run,p_p,N_x,N_y, max_it, itol, p_its, self.Dt, p_z,p_dblparams, p_intparams)

    #     self.t = t
    #     self.it = it
    #     self.it_store = it_store
    #     self.xy = xy
    #     self.z = z
           
    def save_run(self, file_name = ''):

        np.savez_compressed(file_name, Time = self.Time,
                            X = self.X,
                            Y = self.Y,
                            Z = self.Z,
                            params = self.p,
                            u_run = self.u_run,
                            u_ini = self.u_ini)
        
        return None            

    def load_run(self, file_name = ''):

        results = np.load(file_name)
        self.Time = results['Time']
        self.X = results['X']
        self.Y = results['Y']
        self.Z = results['Z']
        self.p = results['params']
        self.u_ini = results['u_ini']
        self.u_run = results['u_run']

        self.t = self.Time[-1]
        self.xy[:self.N_x] = self.X[-1,:]
        self.xy[self.N_x:] = self.Y[-1,:]
        self.z = self.Z[-1,:]    
        return None
        
        
    def full_jacs_eval(self):
        N_x = self.N_x
        N_y = self.N_y
        N_xy = N_x + N_y
    
        sp_jac_run = self.sp_jac_run
        sp_Fu = self.sp_Fu_run
        sp_Gu = self.sp_Gu_run
        sp_Hx = self.sp_Hx_run
        sp_Hy = self.sp_Hy_run
        sp_Hu = self.sp_Hu_run
        
        x = self.xy[0:N_x]
        y = self.xy[N_x:]
        u = self.u_run
        p = self.p
        Dt = self.Dt
    
        sp_jac_run_eval(sp_jac_run.data,x,y,u,p,Dt)
        
        self.Fx = sp_jac_run[0:N_x,0:N_x]
        self.Fy = sp_jac_run[ 0:N_x,N_x:]
        self.Gx = sp_jac_run[ N_x:,0:N_x]
        self.Gy = sp_jac_run[ N_x:, N_x:]
        
        sp_Fu_run_eval(sp_Fu.data,x,y,u,p,Dt)
        sp_Gu_run_eval(sp_Gu.data,x,y,u,p,Dt)
        sp_H_jacs_run_eval(sp_Hx.data,sp_Hy.data,sp_Hu.data,x,y,u,p,Dt)
        
        self.Fu = sp_Fu
        self.Gu = sp_Gu
        self.Hx = sp_Hx
        self.Hy = sp_Hy
        self.Hu = sp_Hu




def sp_jac_ini_vectors():

    sp_jac_ini_ia = [2, 3, 0, 2, 4, 5, 1, 3, 4, 0, 1, 4, 0, 1]
    sp_jac_ini_ja = [0, 1, 2, 6, 9, 12, 14]
    sp_jac_ini_nia = 6
    sp_jac_ini_nja = 6
    return sp_jac_ini_ia, sp_jac_ini_ja, sp_jac_ini_nia, sp_jac_ini_nja 

def sp_jac_run_vectors():

    sp_jac_run_ia = [2, 3, 0, 2, 4, 1, 3, 4, 0, 1, 4, 0, 1, 5]
    sp_jac_run_ja = [0, 1, 2, 5, 8, 11, 14]
    sp_jac_run_nia = 6
    sp_jac_run_nja = 6
    return sp_jac_run_ia, sp_jac_run_ja, sp_jac_run_nia, sp_jac_run_nja 

def sp_jac_trap_vectors():

    sp_jac_trap_ia = [0, 2, 1, 3, 0, 2, 4, 1, 3, 4, 0, 1, 4, 0, 1, 5]
    sp_jac_trap_ja = [0, 2, 4, 7, 10, 13, 16]
    sp_jac_trap_nia = 6
    sp_jac_trap_nja = 6
    return sp_jac_trap_ia, sp_jac_trap_ja, sp_jac_trap_nia, sp_jac_trap_nja 



import numpy as np
import sympy as sym


def dae(model_name):
    L,G,M,K_d = sym.symbols('L,G,M,K_d', real=True)
    p_x,p_y,v_x,v_y = sym.symbols('p_x,p_y,v_x,v_y', real=True) 
    lam,f_x,theta,u_dummy = sym.symbols('lam,f_x,theta,u_dummy', real=True) 

    dp_x = v_x
    dp_y = v_y
    dv_x = (-2*p_x*lam + f_x - K_d*v_x)/M
    dv_y = (-M*G - 2*p_y*lam - K_d*v_y)/M   

    g_1 = p_x**2 + p_y**2 - L**2 -lam*1e-6
    g_2 = -theta + sym.atan2(p_x,-p_y) + u_dummy

    params_dict = {'L':5.21,'G':9.81,'M':10.0,'K_d':1e-3}  # parameters with default values

    u_ini_dict = {'theta':np.deg2rad(5.0),'u_dummy':0.0}  # input for the initialization problem
    u_run_dict = {'f_x':0,'u_dummy':0.0}                  # input for the running problem, its value is updated 

    sys_dict = {'name':model_name,
                'params_dict':params_dict,
                'f_list':[dp_x,dp_y,dv_x,dv_y],
                'g_list':[g_1,g_2],
                'x_list':[ p_x, p_y, v_x, v_y],
                'y_ini_list':[lam,f_x],
                'y_run_list':[lam,theta],
                'u_ini_dict':u_ini_dict,
                'u_run_dict':u_run_dict,
                'h_dict':{'E_p':M*G*(p_y+L),'E_k':0.5*M*(v_x**2+v_y**2),'f_x':f_x,'lam':lam}} 

    return sys_dict


def test_build_run():
    import pydae.build_cffi as db
    sys_dict = dae()
    bldr = db.builder(sys_dict)
    bldr.build()

    from pydae import ssa
    import pendulum 

    model = pendulum.model()

    M = 30.0  # mass of the bob (kg)
    L = 5.21  # length of the pendulum (m)
    model.ini({'M':M,'L':L,           # parameters setting
            'theta':np.deg2rad(0)  # initial desired angle = 0º
            },-1)                  # here -1 means that -1 is considered as initial gess for
                                    # dynamic and algebraic states

    model.report_x()  # obtained dynamic states
    model.report_y()  # obtained algebraic states
    model.report_z()  # obtained outputs
    model.report_u()  # obtained algebraic states (theta is both state and output; f_x is both input and output)
    model.report_params()  # considered parameters

if __name__ == '__main__':

    test_build_run()








# -*- coding: utf-8 -*-
"""
Created on Thu August 10 23:52:55 2022

@author: jmmauricio
"""

import numpy as np
import sympy as sym


def agov1(dae,syn_data,name,bus_name):
    '''
    Governor AGOV1 
    
    '''
    gov_data = syn_data['gov']

    
    # inpunts
    omega = sym.Symbol(f"omega_{name}", real=True)
    p_c = sym.Symbol(f"p_c_{name}", real=True) 
    p_g = sym.Symbol(f"p_g_{name}", real=True)
    p_agc = sym.Symbol("p_agc", real=True)
    p_r = sym.Symbol(f"p_r_{name}", real=True)
    
    # dynamic states
    x_gov_1 = sym.Symbol(f"x_gov_1_{name}", real=True)
    x_gov_2 = sym.Symbol(f"x_gov_2_{name}", real=True)  
    xi_imw  = sym.Symbol(f"xi_imw_{name}", real=True)  

    # algebraic states
    p_m = sym.Symbol(f"p_m_{name}", real=True)  
    p_m_ref = sym.Symbol(f"p_m_ref_{name}", real=True)  

    # parameters
    T_1 = sym.Symbol(f"T_gov_1_{name}", real=True)  # 1
    T_2 = sym.Symbol(f"T_gov_2_{name}", real=True)  # 2
    T_3 = sym.Symbol(f"T_gov_3_{name}", real=True)  # 10
    Droop = sym.Symbol(f"Droop_{name}", real=True)  # 0.05
    K_imw = sym.Symbol(f"K_imw_{name}", real=True)  # 0.0
    K_sec = sym.Symbol(f"K_sec_{name}", real=True)  # 0.0    
    omega_ref = sym.Symbol(f"omega_ref_{name}", real=True)
    
    # auxiliar

    # differential equations
    dx_gov_1 =   (p_m_ref - x_gov_1)/T_1
    dx_gov_2 =   (x_gov_1 - x_gov_2)/T_3
    dxi_imw =   K_imw*(p_c - p_g) - 1e-6*xi_imw

    g_p_m_ref  = -p_m_ref + xi_imw + p_r + K_sec*p_agc - 1/Droop*(omega - omega_ref) 
    g_p_m = (x_gov_1 - x_gov_2)*T_2/T_3 + x_gov_2 - p_m

    
    dae['f'] += [dx_gov_1,dx_gov_2,dxi_imw]
    dae['x'] += [ x_gov_1, x_gov_2, xi_imw]
    dae['g'] += [g_p_m_ref,g_p_m]
    dae['y_ini'] += [  p_m_ref,  p_m]  
    dae['y_run'] += [  p_m_ref,  p_m] 
    dae['params_dict'].update({str(Droop):gov_data['Droop']})
    dae['params_dict'].update({str(T_1):gov_data['T_1']})
    dae['params_dict'].update({str(T_2):gov_data['T_2']})
    dae['params_dict'].update({str(T_3):gov_data['T_3']})
    dae['params_dict'].update({str(K_imw):gov_data['K_imw']})
    dae['params_dict'].update({str(K_sec):syn_data['K_sec']})
    dae['params_dict'].update({str(omega_ref):1.0})

    dae['u_ini_dict'].update({str(p_c):gov_data['p_c']})
    dae['u_run_dict'].update({str(p_c):gov_data['p_c']})
    dae['u_ini_dict'].update({str(p_r):0.0})
    dae['u_run_dict'].update({str(p_r):0.0})


import numpy as np
import json

def get_v(grid,bus_name,type='V_an_m'):
    '''
    Return desired voltages values as defined by argument type.    

    Parameters
    ----------
    grid : pydae object
        Pydae object modelling a grid.

    bus_name : string
        Name of the bus.

    type : string, optional
        Definition of the required voltage.
        V_an_m -> Phase to neutral voltage module for phase a
        V_an -> Phase to neutral voltage complex value for phase a
        U_ab_m -> Phase a to b voltage module

    Returns
    -------
    Real or complex value, also can be a numpy array, depending on the type argument

    Example
    -------

    '''
    
    a2n = {'a':0,'b':1,'c':2,'n':3}

    if type in ['V_an_m','V_bn_m','V_cn_m','V_an','V_bn','V_cn','V_ng','V_ng_m']:
            
        v_n_r,v_n_i =  grid.get_mvalue([f"V_{bus_name}_3_r",f"V_{bus_name}_3_i"])
        v_n_g = v_n_r + 1j*v_n_i

        if type == 'V_ng':  # neutral to ground voltage
            return v_n_g

        if type == 'V_ng_m':
            return np.abs(v_n_g)
        
        ph = a2n[type[2]]
        v_ph_r,v_ph_i =  grid.get_mvalue([f"V_{bus_name}_{ph}_r",f"V_{bus_name}_{ph}_i"])
        v_ph_g = v_ph_r + 1j*v_ph_i

        if type[-1] == 'm':
            return np.abs(v_ph_g - v_n_g)
        else:
            return v_ph_g - v_n_g

    if type in ['V_abcn_m','V_abcn']:
        v_list = []
        v_n_r,v_n_i =  grid.get_mvalue([f"V_{bus_name}_3_r",f"V_{bus_name}_3_i"])
        v_n_g = v_n_r + 1j*v_n_i

        for it in range(3):

            v_ph_r,v_ph_i =  grid.get_mvalue([f"V_{bus_name}_{it}_r",f"V_{bus_name}_{it}_i"])
            v_ph_g = v_ph_r + 1j*v_ph_i

            if type[-1] == 'm':
                v_list += [np.abs(v_ph_g - v_n_g)]
            else:
                v_list += [v_ph_g - v_n_g]
            
        return np.array(v_list)

    if type in ['U_ab_m','U_bc_m','U_ca_m','U_ba_m','U_cb_m','U_ac_m']:
        ph_1,ph_2 = a2n[type[2]],a2n[type[3]]
        v_1_r,v_1_i =  grid.get_mvalue([f"V_{bus_name}_{ph_1}_r",f"V_{bus_name}_{ph_1}_i"])
        v_2_r,v_2_i =  grid.get_mvalue([f"V_{bus_name}_{ph_2}_r",f"V_{bus_name}_{ph_2}_i"])
        return np.abs((v_1_r - v_2_r) + 1j*(v_1_i - v_2_i))
            
    
def get_i(grid,bus_j,bus_k,type='I_a_m'):
    '''
    Return desired currents values as defined by argument type.    

    Parameters
    ----------
    grid : pydae object
        Pydae object modelling a grid.

    bus_name : string
        Name of the bus.

    type : string, optional
        Definition of the required voltage.
        I_a_m -> Phase current module

    Returns
    -------
    Real or complex value, also can be a numpy array, depending on the type argument

    Example
    -------

    '''
    
    a2n = {'a':0,'b':1,'c':2,'n':3}

    if type in ['I_a_m','I_b_m','I_c_m','I_n_m','I_a','I_b','I_c','I_n']:
        
        ph = a2n[type[2]]
        i_r,i_i =  grid.get_mvalue([f"i_l_{bus_j}_{ph}_{bus_k}_{ph}_r",f"i_l_{bus_j}_{ph}_{bus_k}_{ph}_i"])
        i = i_r + 1j*i_i

        if type[-1] == 'm':
            return np.abs(i)
        else:
            return i
        

def get_power(grid,bus_j,bus_k,type='S_a'): 
    '''
    Return desired complex powers.    

    Parameters
    ----------
    grid : pydae object
        Pydae object modelling a grid.

    bus_name : string
        Name of the bus.

    type : string, optional
        Definition of the required voltage.
        S_a_m -> Phase aparent power

    Returns
    -------
    Real or complex value, also can be a numpy array, depending on the type argument

    Example
    -------

    get_power(model,'B2','B3',type='S_c').real 

    '''
    if type in ['S_a_m','S_b_m','S_c_m','S_a','S_b','S_c']:

        ph = type[2]
        V_phn = get_v(grid,bus_j,type=f'V_{ph}n')
        I_ph = get_i(grid,bus_j,bus_k,type=f'I_{ph}')
        S_ph = V_phn*np.conjugate(I_ph)

        if type[-1] == 'm':
            return np.abs(S_ph)
        else:
            return S_ph
        
    if type in ['S_m','S']:
        S_ph = 0.0

        for ph in ['a','b','c']:
            V_phn = get_v(grid,bus_j,type=f'V_{ph}n')
            I_ph = get_i(grid,bus_j,bus_k,type=f'I_{ph}')
            S_ph += V_phn*np.conjugate(I_ph)

        if type[-1] == 'm':
            return np.abs(S_ph)
        else:
            return S_ph